import sys
import datetime
import os.path
import hashlib
import numpy as np

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self
from typing import Any

from pydicom import dcmread
from warnings import warn
from copy import deepcopy

from mirp._data_import.generic_file import ImageFile, MaskFile
from mirp._data_import.utilities import supported_image_modalities, stacking_dicom_image_modalities, \
    supported_mask_modalities, get_pydicom_meta_tag, convert_dicom_time, has_pydicom_meta_tag


class ImageDicomFile(ImageFile):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # These are set using the 'complete' method.
        self.series_instance_uid: None | str = None
        self.frame_of_reference_uid: None | str = None
        self.sop_instance_uid: None | str = None

    def is_stackable(self, stack_images: str):
        """
        Is the image potentially stackable?
        :param stack_images: One of auto, yes or no. Ignored for DICOM imaging, as stackability can be determined
        from metadata.
        :return: boolean value, here true.
        """
        return True

    def get_identifiers(self, as_hash=False) -> dict[str, Any] | bytes:
        """
        General identifiers for images. Compared to other
        :return: a dictionary with identifiers.
        """

        identifier_data = dict({
            "modality": [self.modality],
            "file_type": [self.file_type],
            "sample_name": [self.get_sample_name(allow_none=False)],
            "dir_path": [self.get_dir_path(allow_none=False)],
            "series_instance_uid": [self.series_instance_uid],
            "frame_of_reference_uid": [self.frame_of_reference_uid]
        })

        if as_hash:
            return hashlib.sha256(str(identifier_data).encode(), usedforsecurity=False).digest()
        else:
            return identifier_data

    def check(self, raise_error=False, remove_metadata=True) -> bool:

        # Metadata needs to be read from files, and should therefore be skipped if not available.
        if self.file_path is None:
            return True

        # Checks requires metadata.
        self.load_metadata(limited=True)

        # Perform general checks.
        if not super().check(raise_error=raise_error):
            if remove_metadata:
                self.remove_metadata()
            return False

        if remove_metadata:
            self.remove_metadata()

        return True

    def _check_modality(self, raise_error: bool) -> bool:
        dicom_modality = get_pydicom_meta_tag(dcm_seq=self.image_metadata, tag=(0x0008, 0x0060), tag_type="str")

        # Check for ADC images
        if self._check_is_mr_adc():
            dicom_modality = "adc"
        if self._check_is_mr_dce():
            dicom_modality = "dce"

        support_modalities = supported_image_modalities(self.modality)
        if dicom_modality.lower() not in support_modalities:
            if raise_error:
                raise ValueError(
                    f"The current DICOM file does not have the expected modality. "
                    f"Found: {dicom_modality.lower()}. Expected: one of {', '.join(support_modalities)}. "
                    f"[{self.describe_self()}]"
                )

            return False

        return True

    def _check_sample_name(self, raise_error: bool) -> bool:
        if self.sample_name is not None:
            allowed_sample_name = self.sample_name
            if not isinstance(allowed_sample_name, list):
                allowed_sample_name = [allowed_sample_name]

            # Consider the following DICOM elements: study description, series description, patient name,
            # patient id, study id.
            dicom_sample_name = [
                get_pydicom_meta_tag(dcm_seq=self.image_metadata, tag=(0x0008, 0x1030), tag_type="str"),
                get_pydicom_meta_tag(dcm_seq=self.image_metadata, tag=(0x0008, 0x103E), tag_type="str"),
                get_pydicom_meta_tag(dcm_seq=self.image_metadata, tag=(0x0010, 0x0010), tag_type="str"),
                get_pydicom_meta_tag(dcm_seq=self.image_metadata, tag=(0x0010, 0x0020), tag_type="str"),
                get_pydicom_meta_tag(dcm_seq=self.image_metadata, tag=(0x0020, 0x0010), tag_type="str")
            ]

            dicom_sample_name = [
                current_dicom_sample_name for current_dicom_sample_name in dicom_sample_name
                if current_dicom_sample_name is not None
            ]

            if len(dicom_sample_name) > 0:
                matching_sample_name = set(dicom_sample_name).intersection(set(allowed_sample_name))
                if len(matching_sample_name) == 0:
                    if raise_error:
                        raise ValueError(
                            f"The current DICOM file does not have a matching sample name among potential identifiers. "
                            f"Potential identifiers: {', '.join(dicom_sample_name)}; "
                            f"Expected identifiers: {', '.join(allowed_sample_name)}. "
                            f"[{self.describe_self()}]"
                        )
                    else:
                        return False

                return True
        else:
            return True

    def create(self):
        """
        DICOM-files have different routines depending on the modality. These are then dispatched to different classes
        using this method.
        :return: an object of a subclass of ImageDicomFile.
        """

        # Import locally to prevent circular references.
        from mirp._data_import.dicom_file_ct import ImageDicomFileCT
        from mirp._data_import.dicom_file_mr import ImageDicomFileMR
        from mirp._data_import.dicom_file_mr_adc import ImageDicomFileMRADC
        from mirp._data_import.dicom_file_mr_dce import ImageDicomFileMRDCE
        from mirp._data_import.dicom_file_pet import ImageDicomFilePT
        from mirp._data_import.dicom_file_rtdose import ImageDicomFileRTDose
        from mirp._data_import.dicom_file_cr import ImageDicomFileCR
        from mirp._data_import.dicom_file_dx import ImageDicomFileDX
        from mirp._data_import.dicom_file_mg import ImageDicomFileMG
        from mirp._data_import.dicom_multi_frame import ImageDicomMultiFrame

        # Load metadata so that the modality tag can be read.
        self.load_metadata(limited=True)
        modality = get_pydicom_meta_tag(dcm_seq=self.image_metadata, tag=(0x0008, 0x0060), tag_type="str").lower()

        if modality is None:
            raise TypeError(f"Modality attribute could not be obtained from DICOM file. [{self.describe_self()}]")

        if modality == "mr" and self._check_is_mr_adc():
            modality = "adc"
        if modality == "mr" and self._check_is_mr_dce():
            modality = "dce"

        if modality == "ct":
            file_class = ImageDicomFileCT
        elif modality == "pt":
            file_class = ImageDicomFilePT
        elif modality == "mr":
            file_class = ImageDicomFileMR
        elif modality == "adc":
            file_class = ImageDicomFileMRADC
        elif modality == "dce":
            file_class = ImageDicomFileMRDCE
        elif modality == "rtdose":
            file_class = ImageDicomFileRTDose
        elif modality == "cr":
            file_class = ImageDicomFileCR
        elif modality == "dx":
            file_class = ImageDicomFileDX
        elif modality == "mg":
            file_class = ImageDicomFileMG
        else:
            # This will return a base class, which will fail to pass the modality check.
            return None

        if has_pydicom_meta_tag(dcm_seq=self.image_metadata, tag=(0x5200, 0x9299)) or \
                has_pydicom_meta_tag(dcm_seq=self.image_metadata, tag=(0x5200, 0x9230)):
            file_class = ImageDicomMultiFrame

        # Instantiate subclass, and update using current object.
        # Modality is updated here to reflect the choices made above.
        image = file_class()
        image.update_from_template(template=self)
        image.modality = modality

        # Multi-frame images need additional work.
        if isinstance(image, ImageDicomMultiFrame):
            image = image.create()

        return image

    def update_from_template(self, template: Self):
        if not isinstance(template, ImageDicomFile):
            raise TypeError(
                f"The new class object should inherit from an ImageDicomFile object. Found: {type(template)}"
            )

        # Attributes from the template Image Dicom File. Note that image_data and image_metadata are shallow copies.
        self.file_path = deepcopy(template.file_path)
        self.dir_path = deepcopy(template.dir_path)
        self.sample_name = deepcopy(template.sample_name)
        self.file_name = deepcopy(template.file_name)
        self.modality = deepcopy(template.modality)
        self.image_name = deepcopy(template.image_name)
        self.file_type = deepcopy(template.file_type)
        self.image_data = template.image_data
        self.image_origin = deepcopy(template.image_origin)
        self.image_orientation = deepcopy(template.image_orientation)
        self.image_spacing = deepcopy(template.image_spacing)
        self.image_dimension = deepcopy(template.image_dimension)
        self.frame_of_reference_uid = deepcopy(template.frame_of_reference_uid)
        self.image_metadata = template.image_metadata
        self.is_limited_metadata = deepcopy(template.is_limited_metadata)
        self.series_instance_uid = deepcopy(template.series_instance_uid)
        self.sop_instance_uid = deepcopy(template.sop_instance_uid)
        self.associated_masks = template.associated_masks

    def complete(self, remove_metadata=False, force=False):

        # complete loads metadata.
        super().complete(remove_metadata=False, force=force)

        # Set SOP instance UID.
        if self.sop_instance_uid is None:
            self.load_metadata(limited=True)
            self.sop_instance_uid = get_pydicom_meta_tag(
                dcm_seq=self.image_metadata,
                tag=(0x0008, 0x0018),
                tag_type="str"
            )

        # Set Frame of Reference UID (if any)
        self._complete_frame_of_reference_uid()

        # Set series UID
        if self.series_instance_uid is None:
            self.load_metadata(limited=True)
            self.series_instance_uid = get_pydicom_meta_tag(
                dcm_seq=self.image_metadata,
                tag=(0x0020, 0x000E),
                tag_type="str"
            )

        if remove_metadata:
            self.remove_metadata()

    def _complete_modality(self):
        if self.modality is None:
            self.load_metadata(limited=True)
            self.modality = get_pydicom_meta_tag(dcm_seq=self.image_metadata, tag=(0x0008, 0x0060), tag_type="str")

        if self.modality is None:
            warn(f"Could not ascertain the modality from the DICOM file. [{self.describe_self()}]", UserWarning)

            self.modality = "generic"
        else:
            self.modality = self.modality.lower()

    def _complete_sample_name(self):
        # Set sample name -- note that if sample_name is a (single) string, it is neither replaced nor updated.
        if self.sample_name is None or isinstance(self.sample_name, list):

            # Load relevant metadata.
            self.load_metadata(limited=True)

            # Consider the following DICOM elements:
            # patient id, study id, patient name, series description and study description.
            # These are explicitly ordered by relevance for setting the sample name.
            dicom_sample_name = [
                get_pydicom_meta_tag(dcm_seq=self.image_metadata, tag=(0x0010, 0x0020), tag_type="str"),
                get_pydicom_meta_tag(dcm_seq=self.image_metadata, tag=(0x0020, 0x0010), tag_type="str"),
                get_pydicom_meta_tag(dcm_seq=self.image_metadata, tag=(0x0010, 0x0010), tag_type="str"),
                get_pydicom_meta_tag(dcm_seq=self.image_metadata, tag=(0x0008, 0x1030), tag_type="str"),
                get_pydicom_meta_tag(dcm_seq=self.image_metadata, tag=(0x0008, 0x103E), tag_type="str")
            ]

            dicom_sample_name = [
                current_dicom_sample_name for current_dicom_sample_name in dicom_sample_name
                if current_dicom_sample_name is not None
            ]

            if self.sample_name is None and len(dicom_sample_name) > 0:
                # Set sample name using the most relevant DICOM tag, if no sample names are present.
                self.sample_name = dicom_sample_name[0]

            elif len(self.sample_name) > 0 and len(dicom_sample_name) > 0:
                # Set sample name using the intersection of allowed sample names and DICOM-based names.
                matching_sample_names = set(self.sample_name).intersection(set(dicom_sample_name))
                if len(matching_sample_names) > 0:
                    self.sample_name = list(matching_sample_names)[0]
                else:
                    self.sample_name = None

    def _complete_image_origin(self, force=False, frame_id=None):
        if self.image_origin is None:
            # Origin needs to be determined at the stack-level for slice-based dicom, not for each slice.
            if self.modality in stacking_dicom_image_modalities() and not force:
                return

            # Load relevant metadata.
            self.load_metadata(limited=True)

            origin = get_pydicom_meta_tag(
                dcm_seq=self.image_metadata,
                tag=(0x0020, 0x0032),
                tag_type="mult_float",
                macro_dcm_seq=(0x0020, 0x9113),
                frame_id=frame_id
            )[::-1]
            self.image_origin = tuple(origin)

    def _complete_image_orientation(self, force=False, frame_id=None):
        if self.image_orientation is None:
            # Orientation needs to be determined at the stack-level for slice-based dicom, not for each slice.
            if self.modality in stacking_dicom_image_modalities() and not force:
                return

            # Load relevant metadata.
            self.load_metadata(limited=True)

            orientation: list[float] = get_pydicom_meta_tag(
                dcm_seq=self.image_metadata,
                tag=(0x0020, 0x0037),
                tag_type="mult_float",
                macro_dcm_seq=(0x0020, 0x9116),
                frame_id=frame_id
            )

            # First compute z-orientation.
            # noinspection PyUnreachableCode
            orientation += list(np.cross(orientation[0:3], orientation[3:6]))
            self.image_orientation = np.reshape(orientation[::-1], [3, 3], order="F")

    def _complete_image_spacing(self, force=False, frame_id=None):
        if self.image_spacing is None:
            # Image spacing needs to be determined at the stack-level for slice-based dicom, not for each slice.
            if self.modality in stacking_dicom_image_modalities() and not force:
                return

            # Load relevant metadata.
            self.load_metadata(limited=True)

            # Get pixel-spacing.
            spacing = get_pydicom_meta_tag(
                dcm_seq=self.image_metadata,
                tag=(0x0028, 0x0030),
                tag_type="mult_float",
                macro_dcm_seq=(0x0028, 0x9110),
                frame_id=frame_id
            )

            # First try to get spacing between slices.
            z_spacing = get_pydicom_meta_tag(
                dcm_seq=self.image_metadata,
                tag=(0x0018, 0x0088),
                tag_type="float",
                macro_dcm_seq=(0x0028, 0x9110),
                frame_id=frame_id
            )

            # If spacing between slices is not set, get slice thickness.
            if z_spacing is None:
                z_spacing = get_pydicom_meta_tag(
                    dcm_seq=self.image_metadata,
                    tag=(0x0018, 0x0050),
                    tag_type="float",
                    macro_dcm_seq=(0x0028, 0x9110),
                    frame_id=frame_id
                )

            # If slice thickness is not set, use a default value.
            if z_spacing is None:
                z_spacing = 1.0
            spacing += [z_spacing]

            self.image_spacing = tuple(spacing[::-1])

    def _complete_image_dimensions(self, force=False):
        if self.image_dimension is None:
            # Image dimension needs to be determined at the stack-level for slice-based dicom, not for each slice.
            if self.modality in stacking_dicom_image_modalities() and not force:
                return

            # Load relevant metadata.
            self.load_metadata(limited=True)

            dimensions = tuple([
                1,
                get_pydicom_meta_tag(dcm_seq=self.image_metadata, tag=(0x0028, 0x0010), tag_type="int"),
                get_pydicom_meta_tag(dcm_seq=self.image_metadata, tag=(0x0028, 0x0011), tag_type="int")
            ])

            self.image_dimension = dimensions

    def _complete_frame_of_reference_uid(self):
        if self.frame_of_reference_uid is None:
            self.load_metadata(limited=True)
            self.frame_of_reference_uid = get_pydicom_meta_tag(
                dcm_seq=self.image_metadata,
                tag=(0x0020, 0x0052),
                tag_type="str"
            )

    def associate_with_mask(
            self,
            mask_list,
            association_strategy=None):
        if mask_list is None or len(mask_list) == 0 or association_strategy is None:
            return None

        # Match on frame of reference UID:
        if "frame_of_reference" in association_strategy and self.frame_of_reference_uid is not None:
            matching_mask_list = [
                mask_file for mask_file in mask_list
                if self.frame_of_reference_uid == mask_file.frame_of_reference_uid
            ]

            if len(matching_mask_list) > 0:
                self.associated_masks = matching_mask_list
                return

        return super().associate_with_mask(mask_list=mask_list, association_strategy=association_strategy)

    def load_metadata(self, limited=False, include_image=False):
        if include_image:
            limited = False

        # Limited metadata exists and limited metadata is sufficient.
        if self.image_metadata is not None and self.is_limited_metadata and limited:
            return

        # A full image metadata set exists.
        if self.image_metadata is not None and not self.is_limited_metadata:
            return

        if self.file_path is None or not os.path.exists(self.file_path):
            raise FileNotFoundError(
                f"The image file could not be found at the expected location: {self.file_path}. "
                f"[{self.describe_self()}]")

        if limited:
            dcm = dcmread(
                self.file_path,
                stop_before_pixels=True,
                force=True,
                specific_tags=self._get_limited_metadata_tags()
            )
        else:
            dcm = dcmread(
                self.file_path,
                stop_before_pixels=not include_image,
                force=True
            )

        self.image_metadata = dcm
        self.is_limited_metadata = limited

    def load_data(self, **kwargs):
        raise NotImplementedError(
            "DEV: The load_data method should be specified for subclasses of ImageDicomFile. A generic method does not "
            "exist."
        )

    def load_data_generic(self) -> np.ndarray:
        """
        This is the generic method for loading pixel data from DICOM images that is shared across many modalities.
        :return:
        """
        if self.image_data is not None:
            return self.image_data

        if self.file_path is not None and not os.path.exists(self.file_path):
            raise FileNotFoundError(
                f"The image file could not be found at the expected location: {self.file_path}. [{self.describe_self()}]"
            )

        if self.file_path is None:
            raise ValueError(f"A path to a file was expected, but not present. [{self.describe_self()}]")

        # Load metadata.
        self.load_metadata(include_image=True)
        image_data = self.image_metadata.pixel_array.astype(np.float32)

        # Update data with scale and intercept. These may change per slice.
        rescale_intercept = get_pydicom_meta_tag(dcm_seq=self.image_metadata, tag=(0x0028, 0x1052), tag_type="float", default=0.0)
        rescale_slope = get_pydicom_meta_tag(dcm_seq=self.image_metadata, tag=(0x0028, 0x1053), tag_type="float", default=1.0)
        image_data = image_data * rescale_slope + rescale_intercept

        return image_data

    def set_object_metadata(self):
        self.object_metadata.update(self.export_metadata())

    def export_metadata(self, self_only=False, **kwargs) -> None | dict[str, Any]:
        if not self_only:
            metadata = super().export_metadata()
        else:
            metadata = {}

        self.load_metadata()

        dcm_meta_data = []

        # Scanner type
        scanner_type = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x0008, 0x1090),
            tag_type="str"
        )
        if scanner_type is not None:
            dcm_meta_data += [("scanner_type", scanner_type)]

        # Scanner manufacturer
        manufacturer = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x0008, 0x0070),
            tag_type="str"
        )
        if manufacturer is not None:
            dcm_meta_data += [("manufacturer", manufacturer)]

        # Image type
        image_type = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x0008, 0x0008),
            tag_type="str"
        )
        if image_type is not None:
            dcm_meta_data += [("image_type", image_type)]

        # Patient orientation
        patient_orientation = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x0020, 0x0020),
            tag_type="str"
        )
        if patient_orientation is not None:
            dcm_meta_data += [("patient_orientation", patient_orientation)]

        metadata.update(dict(dcm_meta_data))

        return metadata

    def _check_is_mr_adc(self):
        # Check for ADC images. ADC can sometimes by identified the ADC value in the Image Type (0008,0008) tag,
        # the frame type tag (0008, 9007) or acquisition contrast (0008, 9209) [though this typically should be
        # DIFFUSION].
        image_type = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x0008, 0x0008),
            tag_type="mult_str"
        )
        frame_type = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x0008, 0x9007),
            tag_type="mult_str",
            macro_dcm_seq=(0x0018, 0x9226),
            frame_id=0
        )
        alt_frame_type = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x0008, 0x9007),
            tag_type="mult_str",
            macro_dcm_seq=(0x0040, 0x9092),
            frame_id=0
        )
        acquisition_contrast = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x0008, 0x9209),
            tag_type="str",
            macro_dcm_seq=(0x0018, 0x9226),
            frame_id=0
        )

        if image_type is not None and any(x.lower() == "adc" for x in image_type):
            return True
        elif frame_type is not None and any(x.lower() == "adc" for x in frame_type):
            return True
        elif alt_frame_type is not None and any(x.lower() == "adc" for x in alt_frame_type):
            return True
        elif acquisition_contrast is not None and acquisition_contrast.lower() == "adc":
            return True

        return False

    def _check_is_mr_dce(self):
        # Check for DCE images. DCE can sometimes by identified the DCE value in the Image Type (0008,0008) tag,
        # the frame type tag (0008, 9007) or acquisition contrast (0008, 9209) [though this typically should be
        # DIFFUSION].
        image_type = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x0008, 0x0008),
            tag_type="mult_str"
        )
        frame_type = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x0008, 0x9007),
            tag_type="mult_str",
            macro_dcm_seq=(0x0018, 0x9226),
            frame_id=0
        )
        alt_frame_type = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x0008, 0x9007),
            tag_type="mult_str",
            macro_dcm_seq=(0x0040, 0x9092),
            frame_id=0
        )
        acquisition_contrast = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x0008, 0x9209),
            tag_type="str",
            macro_dcm_seq=(0x0018, 0x9226),
            frame_id=0
        )

        if image_type is not None and any(x.lower() == "dce" for x in image_type):
            return True
        elif frame_type is not None and any(x.lower() == "dce" for x in frame_type):
            return True
        elif alt_frame_type is not None and any(x.lower() == "dce" for x in alt_frame_type):
            return True
        elif acquisition_contrast is not None and acquisition_contrast.lower() == "dce":
            return True

        return False

    @staticmethod
    def _get_limited_metadata_tags():
        # Limited tags are read to populate basic
        return [
            (0x0008, 0x0008),  # image type
            (0x0008, 0x0018),  # SOP instance UID
            (0x0008, 0x0060),  # modality
            (0x0008, 0x1030),  # study description
            (0x0008, 0x103E),  # series description
            (0x0010, 0x0010),  # patient name
            (0x0010, 0x0020),  # patient id
            (0x0018, 0x0050),  # slice thickness
            (0x0018, 0x9087),  # diffusion b-value
            (0x0020, 0x000E),  # series instance UID
            (0x0020, 0x0010),  # study id
            (0x0020, 0x0032),  # origin
            (0x0020, 0x0037),  # orientation
            (0x0020, 0x0052),  # frame of reference UID
            (0x0028, 0x0008),  # number of frames
            (0x0028, 0x0010),  # pixel rows
            (0x0028, 0x0011),  # pixel columns
            (0x0028, 0x0030),  # pixel spacing
            (0x3004, 0x000C),  # grid frame offset vector
            (0x5200, 0x9229),  # shared functional groups sequence
            (0x5200, 0x9230)   # per-frame functional groups sequence
        ]

    def _get_acquisition_start_time(self) -> datetime.datetime:
        self.load_metadata()

        # Start of image acquisition. Prefer Acquisition Datetime (0x0008, 0x002A).
        acquisition_ref_time = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x0008, 0x002A),
            tag_type="str"
        )
        acquisition_ref_time = convert_dicom_time(datetime_str=acquisition_ref_time)

        # Fall back to Acquisition Date (0x0008, 0x002A) and Acquisition Time (0x0008, 0x0032).
        if acquisition_ref_time is None:
            acquisition_start_date = get_pydicom_meta_tag(
                dcm_seq=self.image_metadata,
                tag=(0x0008, 0x0022),
                tag_type="str"
            )
            acquisition_start_time = get_pydicom_meta_tag(
                dcm_seq=self.image_metadata,
                tag=(0x0008, 0x0032),
                tag_type="str"
            )
            acquisition_ref_time = convert_dicom_time(
                date_str=acquisition_start_date,
                time_str=acquisition_start_time
            )

        # Fall back to Private GE Acquisition DateTime (0x0009, 0x100d).
        if acquisition_ref_time is None:
            acquisition_ref_time = get_pydicom_meta_tag(
                dcm_seq=self.image_metadata,
                tag=(0x0009, 0x100d),
                tag_type="str"
            )
            acquisition_ref_time = convert_dicom_time(datetime_str=acquisition_ref_time)

        # Fall back to Series Date and Series Time (
        if acquisition_ref_time is None:
            acquisition_start_date = get_pydicom_meta_tag(
                dcm_seq=self.image_metadata,
                tag=(0x0008, 0x0021),
                tag_type="str"
            )
            acquisition_start_time = get_pydicom_meta_tag(
                dcm_seq=self.image_metadata,
                tag=(0x0008, 0x0031),
                tag_type="str"
            )
            acquisition_ref_time = convert_dicom_time(
                date_str=acquisition_start_date,
                time_str=acquisition_start_time
            )

        # Final check.
        if acquisition_ref_time is None:
            raise ValueError(
                f"Acquisition start time cannot be determined from DICOM metadata."
            )

        return acquisition_ref_time

    def _get_export_attributes(self) -> dict[str, Any]:
        attributes = []
        self.load_metadata()

        study_description = get_pydicom_meta_tag(dcm_seq=self.image_metadata, tag=(0x0008, 0x1030), tag_type="str")
        if study_description is not None and study_description != "":
            attributes += [("study_description", study_description)]

        series_description = get_pydicom_meta_tag(dcm_seq=self.image_metadata, tag=(0x0008, 0x103E), tag_type="str")
        if series_description is not None and series_description != "":
            attributes += [("series_description", series_description)]

        # Try to find the acquisition time (which may be equal to series time).
        try:
            acquisition_time = self._get_acquisition_start_time()
        except ValueError:
            acquisition_time = None
        if acquisition_time is not None:
            attributes += [("acquisition_time", acquisition_time)]

        if self.series_instance_uid is not None:
            attributes += [("series_instance_uid", self.series_instance_uid)]
        if self.frame_of_reference_uid is not None:
            attributes += [("frame_of_reference_uid", self.frame_of_reference_uid)]

        parent_attributes = super()._get_export_attributes()
        parent_attributes.update(dict(attributes))

        return parent_attributes


class MaskDicomFile(ImageDicomFile, MaskFile):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def _check_modality(self, raise_error: bool) -> bool:
        dicom_modality = get_pydicom_meta_tag(dcm_seq=self.image_metadata, tag=(0x0008, 0x0060), tag_type="str")
        support_modalities = supported_mask_modalities(self.modality)
        if dicom_modality.lower() not in support_modalities:
            if raise_error:
                raise ValueError(
                    f"The current DICOM file does not have the expected modality. "
                    f"Found: {dicom_modality.lower()}. "
                    f"Expected: {', '.join(support_modalities)}. "
                    f"[{self.describe_self()}]"
                )

            return False

        return True

    def load_data(self, **kwargs):
        raise NotImplementedError(
            "DEV: The load_data method should be specified for subclasses of MaskDicomFile. A generic method does not "
            "exist."
        )

    def create(self):
        """
        DICOM-files have different routines depending on the modality. These are then dispatched to different classes
        using this method.
        :return: an object of a subclass of ImageDicomFile.
        """

        # Import locally to prevent circular references.
        from mirp._data_import.dicom_file_rtstruct import MaskDicomFileRTSTRUCT
        from mirp._data_import.dicom_file_seg import MaskDicomFileSEG

        # Load metadata so that the modality tag can be read.
        self.load_metadata(limited=True)

        modality = get_pydicom_meta_tag(dcm_seq=self.image_metadata, tag=(0x0008, 0x0060), tag_type="str").lower()

        if modality is None:
            raise TypeError(f"Modality attribute could not be obtained from the DICOM file. [{self.describe_self()}]")

        if modality == "rtstruct":
            file_class = MaskDicomFileRTSTRUCT
        elif modality == "seg":
            file_class = MaskDicomFileSEG
        else:
            return None

        mask = file_class(
            file_path=self.file_path,
            dir_path=self.dir_path,
            sample_name=self.sample_name,
            file_name=self.file_name,
            image_modality=self.modality,
            image_name=self.image_name,
            image_file_type=self.file_type,
            image_data=self.image_data,
            image_origin=self.image_origin,
            image_orientation=self.image_orientation,
            image_spacing=self.image_spacing,
            image_dimensions=self.image_dimension,
            roi_name=self.roi_name
        )

        # Set metadata of mask.
        mask.image_metadata = self.image_metadata
        mask.is_limited_metadata = self.is_limited_metadata

        return mask

    def _get_limited_metadata_tags(self):
        tags = super()._get_limited_metadata_tags()

        tags += [
            (0x3006, 0x0020),  # Structure set roi sequence
            (0x0028, 0x0008)  # number of frames
        ]



import datetime
import os.path
import fnmatch
import math
from typing import Any, Iterable
from os.path import split

import numpy as np
import pydicom
from pydicom import FileDataset, Dataset, datadict
from pydicom.tag import Tag
def lookup_modality(modality: None | str) -> list[str]:
    modalities = []
    try:
        modalities += supported_mask_modalities(modality=modality)
    except ValueError:
        pass
    try:
        modalities += supported_image_modalities(modality=modality)
    except ValueError:
        pass

    return modalities


def supported_image_modalities(modality: None | str = None) -> list[str]:

    if isinstance(modality, str):
        modality = modality.lower()

    if modality is None:
        return ["ct", "pt", "mr", "rtdose", "cr", "dx", "mg", "generic", "adc", "dce"]

    elif modality == "ct":
        return ["ct"]

    elif modality in ["pt", "pet"]:
        return ["pt"]

    elif modality in ["mr", "mri"]:
        return ["mr"]

    elif modality in ["adc"]:
        return ["adc"]

    elif modality in ["dce"]:
        return ["dce"]

    elif modality in ["rtdose"]:
        return ["rtdose"]

    elif modality in ["cr", "computed_radiography"]:
        return ["cr"]

    elif modality in ["dx", "digital_xray"]:
        return ["dx"]

    elif modality in ["mg", "mammography", "digital_mammography"]:
        return ["mg"]

    elif modality == "generic":
        return ["generic"]

    else:
        raise ValueError(
            f"Encountered an unknown image modality: {modality}. The following image modalities are supported: "
            f"{', '.join(supported_image_modalities(None))}. The generic modality lacks special default parameters, "
            f"and can always be used.")


def stacking_dicom_image_modalities() -> list[str]:
    return ["ct", "pt", "mr", "adc", "dce"]


def supported_mask_modalities(modality: None | str = None) -> list[str]:

    if isinstance(modality, str):
        modality = modality.lower()

    if modality is None:
        return ["rtstruct", "seg", "generic_mask"]

    elif modality in ["rtstruct"]:
        return ["rtstruct"]

    elif modality in ["seg"]:
        return ["seg"]

    elif modality == "generic_mask":
        return ["generic_mask"]

    else:
        raise ValueError(
            f"Encountered an unknown mask modality: {modality}. The following mask modalities are supported: "
            f"{', '.join(supported_mask_modalities(None))}. The generic modality can always be used.")


def supported_file_types(file_type: None | str = None) -> list[str]:

    if isinstance(file_type, str):
        file_type = file_type.lower()

    if file_type is None:
        return [".dcm", ".nii", ".nii.gz", ".nrrd", ".npy"]

    elif file_type == "dicom":
        return [".dcm"]

    elif file_type == "itk":
        return [".nii", ".nii.gz", ".nrrd"]

    elif file_type == "nifti":
        return [".nii", ".nii.gz"]

    elif file_type == "nrrd":
        return [".nrrd"]

    elif file_type == "numpy":
        return [".npy"]

    else:
        raise ValueError(
            f"Encountered an unknown file type argument: {file_type}. The following file types are supported: "
            f"{', '.join(supported_file_types(None))}.")


def flatten_list(unflattened_list):

    if len(unflattened_list) == 0:
        return unflattened_list

    if isinstance(unflattened_list[0], list):
        return flatten_list(unflattened_list[0]) + flatten_list(unflattened_list[1:])

    return unflattened_list[:1] + flatten_list(unflattened_list[1:])


def bare_file_name(
        x: str | list[str],
        file_extension: str | list[str]
) -> str | list[str]:
    """
    Strips provided extensions from the name of a file.
    :param x: One or more filenames or path to file names.
    :param file_extension: One or more extensions that should be stripped
    :return: One or more filenames from which the extension has been stripped.
    """
    return_list = True
    if isinstance(x, str):
        x = [x]
        return_list = False

    if isinstance(file_extension, str):
        file_extension = [file_extension]

    file_name = [os.path.basename(file_path) for file_path in x]

    for ii, current_file_name in enumerate(file_name):
        for extension in file_extension:
            if current_file_name.endswith(extension):
                file_name[ii] = current_file_name.removesuffix(extension)

                # If a file extension is found, remove it only once -- we want to avoid accidentally stripping the
                # filename more than necessary.
                break

    if return_list:
        return file_name
    else:
        return file_name[0]


def match_file_name(
        x: str | list[str],
        pattern: str | list[str],
        file_extension: None | str | list[str]
) -> bool | list[bool]:
    """
    Determine if any filename matches the provided pattern. fnmatch is used for matching, which allows for wildcards.
    :param x: a string or path that is the filename or a path to the file.
    :param pattern: a string or list of strings that should be tested.
    :param file_extension: None, string or list of strings representing the file extension. If provided, the extension
    is stripped from the filename prior to matching.
    :return: a (list of) boolean value(s). True if any pattern appears in the file name, and False if not.
    """
    return_list = True
    if isinstance(x, str):
        x = [x]
        return_list = False

    if isinstance(pattern, str):
        pattern = [pattern]

    file_name = [os.path.basename(file_path) for file_path in x]

    if file_extension is not None:
        file_name = bare_file_name(file_name, file_extension=file_extension)

    matches = np.zeros(len(file_name), dtype=bool)
    for current_pattern in pattern:
        current_pattern = current_pattern.replace("#", "*")
        matches = np.logical_or(matches, np.array([
            fnmatch.fnmatch(current_file_name, current_pattern)
            for current_file_name in file_name
        ]))

    if return_list:
        return matches
    else:
        return any(matches)


def isolate_sample_name(
        x: str,
        pattern: str,
        file_extenstion: None | str | list[str]
) -> None | str:

    # Pattern should only contain one sample name placeholder (#).
    if pattern.count("#") != 1:
        return None

    x = bare_file_name(x, file_extension=file_extenstion)

    # Determine where the sample name placeholder is compared to other wildcards.
    central_split_id = 0
    for current_character in pattern:
        if current_character == "#":
            break
        elif current_character == "*":
            central_split_id += 1

    pattern = pattern.replace("#", "*")
    if not fnmatch.fnmatch(x, pattern):
        return None

    pattern_split = pattern.split("*")
    # Use the fixed (non-wildcard) characters to reduce the string. This is done by stripping away parts to the left
    # or right of fixed characters based on their position relative to the sample name placeholder.
    for ii in range(0, central_split_id + 1):
        if pattern_split[ii] == "":
            continue
        x = x.split(pattern_split[ii], 1)[1]

    for ii in reversed(range(central_split_id + 1, len(pattern_split))):
        if pattern_split[ii] == "":
            continue
        x = x.rsplit(pattern_split[ii], 1)[0]

    if x == "":
        return None

    return x


def path_to_parts(x: str) -> list[str]:
    """
    Split a path into its components.

    Parameters
    ---------
    x: str
        A string that represents a file path.

    Returns
    -------
    path_parts: list of str
        A list of path components.
    """

    path_parts = []
    x_head = x
    while True:
        x_head, x_tail = split(x_head)
        if x_tail == "":
            path_parts += [x_head]
            break
        path_parts += [x_tail]

    return list(reversed(path_parts))


def dir_structure_contains_directory(
        x: str,
        pattern: str | list[str],
        ignore_dir: None | str | list[str]
) -> bool:
    """
    Identify if a path contains a directory matches any of pattern.
    :param x: a string or a path.
    :param pattern: a pattern that should be fully matched in the path.
    :param ignore_dir: any (partial) path that should be ignored. These are stripped from x prior to pattern matching.
    :return: a boolean value.
    """
    # Split x into parts.
    x = path_to_parts(x)

    # Strip the pattern to be ignored from x, if possible.
    if ignore_dir is not None:
        if not isinstance(ignore_dir, list):
            ignore_dir = [ignore_dir]

        for current_ignore_dir in ignore_dir:
            current_ignore_dir = path_to_parts(current_ignore_dir)

            # Find matching sequential elements.
            match_index: list[None | int] = [None for ii in range(len(current_ignore_dir))]
            for jj, ignore_elem in enumerate(current_ignore_dir):
                if match_index[0] is None:
                    if jj > 0:
                        # Break if the first element is not found at all.
                        break
                    for ii in range(len(x)):
                        if x[ii] == ignore_elem:
                            match_index[0] = ii
                            break
                else:
                    if match_index[jj - 1] + 1 > len(x) - 1:
                        # Break if we would exceed the length of x.
                        break

                    if x[match_index[jj - 1] + 1] == ignore_elem:
                        match_index[jj] = match_index[jj - 1] + 1
                    else:
                        # Each element must be sequential.
                        break

            if not any(match_elem is None for match_elem in match_index):
                x = [x_elem for ii, x_elem in enumerate(x) if ii not in match_index]

    if len(x) == 0 or x == "":
        return False

    if not isinstance(pattern, list):
        pattern = [pattern]

    # Find if any pattern is exactly matched.
    return any(x_elem in pattern for x_elem in x)


def compute_file_distance(
        x: None | str,
        y: None | str
) -> int | Any:
    if x is None or y is None:
        return math.inf

    try:
        common_path = os.path.commonpath([x, y])
    except ValueError:
        return math.inf

    # Split to
    split_common_path = path_to_parts(common_path)
    split_x = path_to_parts(x)

    return len(split_x) - len(split_common_path)


def parse_image_correction(
        dcm_seq: pydicom.Dataset,
        tag: tuple[hex, hex],
        correction_abbr: str
) -> bool:
    """
    Parses image correction information. Indicates whether a specific type of PET correction was applied based on
    available information.

    Parameters
    ----------
    dcm_seq: pydicom.Dataset
        Set of DICOM metadata.

    tag: tag for image correction
        Tag for reading image correction from the enhanced PET set of tags.

    correction_abbr:
        Abbreviation for the specific correction in the image_corrections list.

    Returns
    -------
    bool, optional
    """
    image_corrections = get_pydicom_meta_tag(
        dcm_seq=dcm_seq,
        tag=(0x0028, 0x0051),
        tag_type="str"
    )
    if image_corrections is not None:
        image_corrections = image_corrections.replace(
            " ", "").replace(
            "[", "").replace(
            "]", "").replace(
            "\'", "").split(sep=",")

    # Read from enhanced PET.
    is_corrected = get_pydicom_meta_tag(dcm_seq=dcm_seq, tag=tag, tag_type="str")
    if is_corrected is None and image_corrections is None:
        is_corrected = "NO"
    elif is_corrected is None and correction_abbr in image_corrections:
        is_corrected = "YES"
    elif is_corrected is None and correction_abbr not in image_corrections:
        is_corrected = "NO"
    else:
        pass

    return is_corrected == "YES"


def convert_dicom_time(
        datetime_str: None | str = None,
        date_str: None | str = None,
        time_str: None | str = None
) -> None | datetime.datetime:
    """
    Converts DICOM date, time or datetime string to a datetime.datetime object to facilitate use in Python.

    Parameters
    ----------
    datetime_str: str, optional
        Datetime string extract from a DICOM tag.

    date_str: str, optional
        Date string extracted from a DICOM tag.

    time_str: str, optional
        Time string extracted from a DICOM tag.

    Returns
    -------
    datetime.datetime
    """

    if datetime_str is None and (date_str is None or time_str is None):
        # No reference time can be established
        ref_time = None

    elif datetime_str is not None:
        # Single datetime string provided
        year = int(datetime_str[0:4])
        month = int(datetime_str[4:6])
        day = int(datetime_str[6:8])
        hour = int(datetime_str[8:10])
        minute = int(datetime_str[10:12])
        second = int(datetime_str[12:14])
        if len(datetime_str) > 14:
            microsecond = int(round(float(datetime_str[14:]) * 1000))
        else:
            microsecond = 0

        ref_time = datetime.datetime(
            year=year,
            month=month,
            day=day,
            hour=hour,
            minute=minute,
            second=second,
            microsecond=microsecond
        )

    else:
        # Separate date and time strings provided
        year = int(date_str[0:4])
        month = int(date_str[4:6])
        day = int(date_str[6:8])
        hour = int(time_str[0:2])
        minute = int(time_str[2:4])
        second = int(time_str[4:6])
        if len(time_str) > 6:
            microsecond = int(round(float(time_str[6:]) * 1000))
        else:
            microsecond = 0

        ref_time = datetime.datetime(
            year=year,
            month=month,
            day=day,
            hour=hour,
            minute=minute,
            second=second,
            microsecond=microsecond
        )

    return ref_time


def get_pydicom_meta_tag(
        dcm_seq: pydicom.Dataset,
        tag: tuple[hex, hex],
        tag_type: None | str = None,
        default: Any = None,
        macro_dcm_seq: None | tuple[hex, hex] = None,
        frame_id: None | int = None,
        test_tag: bool = False
) -> Any:
    """
    Extract parameter from DICOM metadata.

    Parameters
    ----------
    dcm_seq: pydicom.Dataset
        DICOM sequence from which the parameter should be read.

    tag: tuple of hex
        Hexadecimal values representing the DICOM parameter tag.

    tag_type: str, optional
        Type to which the parameter should be converted.

    default: any, optional, default: None
        Default value to be used in absence of any value from the DICOM metadata.

    macro_dcm_seq: tuple of hex
        Hexadecimal value for the macro sequence within shared or per-frame functional groups.

    frame_id: int
        Index of the frame of interest for tags in per-frame functional groups.

    test_tag: bool, optional, default: False
        Determine whether a tag exists.

    Returns
    -------
    Any
        Value of the parameter read for from the DICOM metadata.
    """
    # Reads dicom tag

    # Initialise with default
    tag_value = default

    while True:
        # Tags are searched in the following order:
        # 1. General header
        # 2. Frame functional group (if frame id is provided).
        # 3. Shared functional group

        # Tag in general header
        try:
            tag_value = dcm_seq[tag].value
            break
        except KeyError:
            pass

        # Tag in frame functional group [0x5200, 0x9230].
        # First test in the macro sequence, if provided. By definition, these sequences only contain a single set of
        # tags.
        if frame_id is not None and macro_dcm_seq is not None:
            try:
                tag_value = dcm_seq[(0x5200, 0x9230)][frame_id][macro_dcm_seq][0][tag].value
                break
            except KeyError:
                pass
        # If not found, test whether the tag is found in the general frame functional group instead of the macro
        # sequence.
        if frame_id is not None:
            try:
                tag_value = dcm_seq[(0x5200, 0x9230)][frame_id][tag].value
                break
            except KeyError:
                pass

        # Tag in shared functional group [0x5200, 0x9229].
        # First test in the macro sequence, if provided. By definition, these sequences only contain a single set of
        # tags.
        if macro_dcm_seq is not None:
            try:
                tag_value = dcm_seq[(0x5200, 0x9229)][0][macro_dcm_seq][0][tag].value
                break
            except KeyError:
                pass
        # If not found, test whether the tag is found in the general frame functional group instead of the macro
        # sequence.
        try:
            tag_value = dcm_seq[(0x5200, 0x9229)][0][tag].value
            break
        except KeyError:
            pass

        # If the look-up gets to this point, the tag is absent.
        if test_tag:
            return False

        # This point in the loop is only reached if the tag could not be found.
        break

    if test_tag:
        return True

    if isinstance(tag_value, bytes):
        tag_value = tag_value.decode("ASCII")

    # Find empty entries
    if tag_value is not None:
        if tag_value == "":
            tag_value = default

    # Cast to correct type (meta tags are usually passed as strings)
    if tag_value is not None:

        # String
        if tag_type == "str":
            tag_value = str(tag_value)

        # Float
        elif tag_type == "float":
            tag_value = float(tag_value)

        # Multiple floats
        elif tag_type == "mult_float":
            tag_value = [float(str_num) for str_num in tag_value]

        # Integer
        elif tag_type == "int":
            tag_value = int(tag_value)

        # Multiple floats
        elif tag_type == "mult_int":
            tag_value = [int(str_num) for str_num in tag_value]

        # Boolean
        elif tag_type == "bool":
            tag_value = bool(tag_value)

        elif tag_type == "mult_str":
            tag_value = list(tag_value)

        else:
            raise ValueError(f"The tag type was not recognised: {tag_type}")

    return tag_value


def has_pydicom_meta_tag(
        dcm_seq: FileDataset | Dataset,
        tag: tuple[hex, hex],
        macro_dcm_seq: None | tuple[hex, hex] = None,
        frame_id: None | int = None
) -> bool:

    return get_pydicom_meta_tag(
        dcm_seq=dcm_seq,
        tag=tag,
        macro_dcm_seq=macro_dcm_seq,
        frame_id=frame_id,
        test_tag=True
    )


def set_pydicom_meta_tag(
        dcm_seq: FileDataset | Dataset,
        tag: tuple[hex, hex] | list[hex],
        value: Any,
        force_vr: None | str = None):
    # Check tag
    if isinstance(tag, tuple):
        tag = Tag(tag[0], tag[1])

    elif isinstance(tag, list):
        tag = Tag(tag[0], tag[2])

    else:
        raise TypeError(f"Metadata tag {tag} is not a pydicom Tag, or can be parsed to one.")

    # Read the default VR information for non-existent tags.
    vr, vm, name, is_retired, keyword = datadict.get_entry(tag)

    if vr == "DS":
        # Decimal string (16-byte string representing decimal value)
        if isinstance(value, Iterable):
            value = [f"{x:.16f}"[:16] for x in value]
        else:
            value = f"{value:.16f}"[:16]

    if tag in dcm_seq and force_vr is None:
        # Update the value of an existing tag.
        dcm_seq[tag].value = value

    elif force_vr is None:
        # Add a new entry.
        dcm_seq.add_new(tag=tag, VR=vr, value=value)

    else:
        # Add a new entry
        dcm_seq.add_new(tag=tag, VR=force_vr, value=value)



import os.path
import numpy as np

from typing import Any

from mirp._data_import.generic_file import ImageFile
from mirp._data_import.dicom_file import ImageDicomFile
from mirp._data_import.utilities import get_pydicom_meta_tag


class ImageDicomPlanarImage(ImageDicomFile):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def is_stackable(self, stack_images: str):
        # Multi-frame images might be actually be stackable (concatenated), but ignore that for now.
        return False

    def _complete_image_origin(self, force=False, frame_id=None):
        if self.image_origin is None:
            self.image_origin = tuple([0.0, 0.0, 0.0])

    def _complete_image_orientation(self, force=False, frame_id=None):
        if self.image_orientation is None:
            self.image_orientation = np.array([[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]])

    def _complete_image_spacing(self, force=False, frame_id=None):
        if self.image_spacing is None:
            # Load relevant metadata.
            self.load_metadata(limited=True)

            # Get pixel-spacing.
            spacing = get_pydicom_meta_tag(
                dcm_seq=self.image_metadata,
                tag=(0x0028, 0x0030),
                tag_type="mult_float",
                macro_dcm_seq=(0x0028, 0x9110),
                frame_id=frame_id
            )
            # Fall-back option if no spacing is provided.
            if spacing is None:
                spacing = [1.0, 1.0]

            # 2D images don't have a depth. The z-dimension is length 1 by convention.
            spacing += [1.0]

            self.image_spacing = tuple(spacing[::-1])

    def _complete_image_dimensions(self, force=False):
        if self.image_dimension is None:
            # Load relevant metadata.
            self.load_metadata(limited=True)

            dimensions = tuple([
                1,
                get_pydicom_meta_tag(dcm_seq=self.image_metadata, tag=(0x0028, 0x0010), tag_type="int"),
                get_pydicom_meta_tag(dcm_seq=self.image_metadata, tag=(0x0028, 0x0011), tag_type="int")
            ])

            self.image_dimension = dimensions

    def load_data(self, **kwargs):
        if self.image_data is not None:
            return self.image_data

        if self.file_path is not None and not os.path.exists(self.file_path):
            raise FileNotFoundError(
                f"The image file could not be found at the expected location: {self.file_path}. [{self.describe_self()}]"
            )

        if self.file_path is None:
            raise ValueError(f"A path to a file was expected, but not present. [{self.describe_self()}]")

        # Load metadata.
        self.load_metadata(include_image=True)
        image_data = self.image_metadata.pixel_array.astype(np.float32)

        # Rescaling and intercept are not required for x-ray images. However, since the pixel data is 2D, we need to
        # add a dimension.
        self.image_data = np.expand_dims(image_data, axis=0)

    def export_metadata(self, self_only=False, **kwargs) -> None | dict[str, Any]:
        if not self_only:
            metadata = super().export_metadata()
        else:
            metadata = {}

        self.load_metadata()

        dcm_meta_data = []

        # Peak kilo voltage output
        kvp = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x0018, 0x0060),
            tag_type="float",
            macro_dcm_seq=(0x0018, 0x9325)
        )
        if kvp is not None:
            dcm_meta_data += [("kvp", kvp)]

        # Tube current in mA
        tube_current = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x0018, 0x1151),
            tag_type="float"
        )
        if tube_current is not None:
            dcm_meta_data += [("tube_current", tube_current)]

        # Exposure time in milliseconds
        exposure_time = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x0018, 0x1150),
            tag_type="float"
        )
        if exposure_time is not None:
            dcm_meta_data += [("exposure_time", exposure_time)]

        # Radiation exposure in mAs
        exposure = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x0018, 0x1152),
            tag_type="float"
        )
        if exposure is not None:
            dcm_meta_data += [("exposure", exposure_time)]

        metadata.update(dict(dcm_meta_data))

        return metadata



import copy
import fnmatch
import hashlib
import itertools
import os
import os.path
import re
import sys
import warnings

import numpy as np

from typing import Any
if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

from mirp._images.base_image import BaseImage
from mirp._images.generic_image import GenericImage
from mirp._masks.base_mask import BaseMask
from mirp._data_import.utilities import supported_file_types, match_file_name, bare_file_name, compute_file_distance, \
    path_to_parts


class ImageFile(BaseImage):

    def __init__(
            self,
            file_path: None | str = None,
            dir_path: None | str = None,
            sample_name: None | str | list[str] = None,
            file_name: None | str = None,
            image_name: None | str = None,
            image_modality: None | str = None,
            image_file_type: None | str = None,
            image_data: None | np.ndarray = None,
            image_origin: None | tuple[float, float, float] = None,
            image_orientation: None | np.ndarray = None,
            image_spacing: None | tuple[float, float, float] = None,
            image_dimensions: None | tuple[int, int, int] = None,
            **kwargs
    ):

        super().__init__(
            image_modality=image_modality,
            image_origin=image_origin,
            image_orientation=image_orientation,
            image_spacing=image_spacing,
            image_dimensions=image_dimensions,
            **kwargs
        )

        self.file_path = file_path
        self.sample_name = sample_name
        self.image_name = image_name

        if isinstance(image_file_type, str):
            image_file_type = image_file_type.lower()
        self.file_type = image_file_type

        # Add image data
        self.image_data = image_data

        # Check incoming image data.
        _ = self._check_image_data()

        self.update_image_data()

        # Check updated image data.
        _ = self._check_image_data()

        # Add metadata
        self.image_metadata = None
        self.is_limited_metadata = False

        # Attempt to set the file name, if this is not externally provided.
        if isinstance(file_path, str) and file_name is None:
            file_name = os.path.basename(file_path)

        # Attempt to set the directory path, if this is not externally provided.
        if isinstance(file_path, str) and dir_path is None:
            dir_path = os.path.dirname(file_path)

        self.file_name = file_name
        self.dir_path = dir_path
        self.associated_masks = None

        # Object metadata that will be exported when converted to native format.
        self.object_metadata: dict[str, str] = dict()

    def is_stackable(self, stack_images: str):
        raise NotImplementedError(
            f"DEV: There is (intentionally) no generic implementation of is_stackable. Please specify "
            f"implementation for subclasses."
        )

    def copy(self):
        return copy.deepcopy(self)

    def on_file_system(self):
        return self.file_path is None or self.dir_path is None

    def get_identifiers(self, as_hash=False) -> dict | bytes:
        """
        General identifiers for images. Note that image_origin is not included, as this should be different for every
        slice in a volume.
        :param as_hash: boolean flag. When true returns a SHA256 hash of the identifier data.
        :return: a dictionary with identifiers.
        """

        identifier_data = dict({
            "modality": [self.modality],
            "file_type": [self.file_type],
            "sample_name": [self.get_sample_name(allow_none=False)],
            "dir_path": [self.get_dir_path(allow_none=False)],
            "image_dimensions": [self.get_image_dimension(as_str=True)],
            "image_spacing": [self.get_image_spacing(as_str=True)],
            "image_orientation": [self.get_image_orientation(as_str=True)]
        })

        if as_hash:
            return hashlib.sha256(str(identifier_data).encode(), usedforsecurity=False).digest()
        else:
            return identifier_data

    def associate_with_mask(
            self,
            mask_list,
            association_strategy: None | set[str] = None
    ):
        import math

        if mask_list is None or len(mask_list) == 0 or association_strategy is None:
            return

        # Match on sample name.
        if "sample_name" in association_strategy and self.sample_name is not None:
            matching_mask_list = [
                mask_file for mask_file in mask_list
                if self.sample_name == mask_file.sample_name
            ]

            if len(matching_mask_list) > 0:
                self.associated_masks = matching_mask_list
                return

        # Match on path distance. Then masks are filtered based on distance to the image, with masks with minimum
        # distance to the image being kept.
        if "file_distance" in association_strategy and self.dir_path is not None:
            file_distance = [
                compute_file_distance(self.dir_path, mask_file.dir_path)
                for mask_file in mask_list
            ]

            if any(math.isfinite(current_distance) for current_distance in file_distance):
                min_distance = min(file_distance)
                matching_mask_list = [
                    mask_file for ii, mask_file in enumerate(mask_list)
                    if file_distance[ii] == min_distance
                ]

                self.associated_masks = matching_mask_list
                return

        # Match on file name similarity. The image is matched against the most similar mask.
        if "file_name_similarity" in association_strategy and self.file_name is not None:
            ...

        return

    def set_sample_name(self, sample_name: str):

        self.sample_name = sample_name

    def get_sample_name(self, allow_none = True):
        if self.sample_name is None:
            if allow_none:
                return None
            else:
                return "unset_sample_name"

        return self.sample_name

    def get_dir_path(self, allow_none = True):
        if self.dir_path is None:
            if allow_none:
                return None
            else:
                return "unset_dir_path"

        return self.dir_path

    def get_file_name(self, allow_none = True):
        if self.file_name is None:
            if allow_none:
                return None
            else:
                return "unset_file_name"

        return self.file_name

    def get_image_dimension(self, as_str: bool = False):
        if not as_str:
            return self.image_dimension

        if self.image_dimension is None:
            return "unset_image_dimension"

        return str(self.image_dimension)

    def get_image_spacing(self, as_str: bool = False):
        if not as_str:
            return self.image_spacing

        if self.image_spacing is None:
            return "unset_image_spacing"

        return str(self.image_spacing)

    def get_image_origin(self, as_str: bool = False):
        if not as_str:
            return self.image_origin

        if self.image_origin is None:
            return "unset_image_origin"

        return str(self.image_origin)

    def get_image_orientation(self, as_str: bool = False):
        if not as_str:
            return self.image_orientation

        if self.image_orientation is None:
            return "unset_image_orientation"

        return str(np.ravel(self.image_orientation))

    def set_modality(self, modality: None | str):
        from mirp._data_import.utilities import supported_image_modalities
        if modality is None:
            return

        if not isinstance(modality, str):
            raise ValueError(f"modality is expected to be a character string. Found: {modality} [{self.describe_self()}]")

        if modality == "generic":
            raise ValueError(f"modality cannot be 'generic' [{self.describe_self()}]")

        modality = supported_image_modalities(modality)
        if self.modality is None or self.modality == "generic":
            self.modality = modality[0]

    def create(self):
        # Import locally to avoid potential circular references.
        from mirp._data_import.dicom_file import ImageDicomFile
        from mirp._data_import.itk_file import ImageITKFile
        from mirp._data_import.numpy_file import ImageNumpyFile

        file_extensions = supported_file_types(file_type=self.file_type)

        if any(self.file_path.lower().endswith(ii) for ii in file_extensions) and\
                any(self.file_path.lower().endswith(ii) for ii in supported_file_types("dicom")):

            file_class = ImageDicomFile
            file_type = "dicom"

        elif any(self.file_path.lower().endswith(ii) for ii in file_extensions) and\
                any(self.file_path.lower().endswith(ii) for ii in supported_file_types("itk")):
            if any(self.file_path.lower().endswith(ii) for ii in supported_file_types("nifti")):
                file_type = "nifti"
            elif any(self.file_path.lower().endswith(ii) for ii in supported_file_types("nrrd")):
                file_type = "nrrd"
            else:
                raise ValueError(f"DEV: specify file_type")

            file_class = ImageITKFile

        elif any(self.file_path.lower().endswith(ii) for ii in file_extensions) and\
                any(self.file_path.lower().endswith(ii) for ii in supported_file_types("numpy")):

            file_class = ImageNumpyFile
            file_type = "numpy"

        else:
            raise NotImplementedError(
                f"The provided image type is not implemented: {self.file_type} [{self.describe_self()}]"
            )

        image_file = file_class(
            file_path=self.file_path,
            dir_path=self.dir_path,
            sample_name=self.sample_name,
            file_name=self.file_name,
            image_name=self.image_name,
            image_modality=self.modality,
            image_file_type=file_type,
            image_data=self.image_data,
            image_origin=self.image_origin,
            image_orientation=self.image_orientation,
            image_spacing=self.image_spacing,
            image_dimensions=self.image_dimension
        ).create()

        return image_file

    def check(self, raise_error=False, remove_metadata=False):

        # Check image data first.
        image_check = self._check_image_data(raise_error=raise_error)
        if not image_check:
            return False

        # Check if file_path is set. Otherwise, none of the generic checks below can be used.
        if self.file_path is None:
            return True

        # Dispatch to subclass based on file_path.
        allowed_file_extensions = supported_file_types(self.file_type)

        # Check that the file type is correct.
        if not self.file_path.lower().endswith(tuple(allowed_file_extensions)):
            if raise_error:
                raise ValueError(
                    f"The file type does not correspond to a known, implemented image type: {self.file_type}. "
                    f"[{self.describe_self()}]"
                )

            return False

        # Check that the file exists.
        if not os.path.exists(self.file_path):
            if raise_error:
                raise FileNotFoundError(
                    f"The image file could not be found at the expected location: {self.file_path} "
                    f"[{self.describe_self()}]"
                )

            return False

        # Check that the file name contains image_name.
        if self.image_name is not None:
            if not match_file_name(self.file_path, pattern=self.image_name, file_extension=allowed_file_extensions):
                if raise_error:
                    raise ValueError(
                        f"The file name of the image file {os.path.basename(self.file_path)} does not match "
                        f"the expected pattern: {self.image_name} [{self.describe_self()}]"
                    )

            return False

        # Check modality.
        modality_check = self._check_modality(raise_error=raise_error)
        if not modality_check:
            return False

        # Check sample name.
        sample_name_check = self._check_sample_name(raise_error=raise_error)
        if not sample_name_check:
            return False

        return True

    def _check_image_data(self, raise_error=False) -> bool:
        if self.image_data is None:
            return True

        # Check that image_data has the expected type.
        if not isinstance(self.image_data, np.ndarray):
            if raise_error:
                raise TypeError(
                    f"The image_data argument expects None or a numpy ndarray. Found object with class "
                    f"{type(self.image_data).name} [{self.describe_self()}]"
                )
            else:
                return False

        # Check that image_data has up to 3 dimensions.
        data_shape = self.image_data.shape
        if not 1 <= len(data_shape) <= 3:
            if raise_error:
                raise ValueError(
                    f"MIRP supports image data up to 3 dimensions. The current numpy array has a dimension of "
                    f"{len(data_shape)} ({data_shape}). [{self.describe_self()}]"
                )
            else:
                return False

        # Check that the shape of image_data matches that of image_dimensions.
        if self.image_dimension is not None and not np.array_equal(self.image_dimension, data_shape):
            if raise_error:
                raise ValueError(
                    f"The shape of the image data itself and the purported shape (image_dimensions) are different. The "
                    f"current numpy array has dimensions ({data_shape}), where ({self.image_dimension}) is expected."
                    f"[{self.describe_self()}]"
                )
            else:
                return False

        # Check that image_origin has the correct number of dimensions.
        if self.image_origin is not None and not len(data_shape) == len(self.image_origin):
            if raise_error:
                raise ValueError(
                    f"The dimensions of the image data itself ({len(data_shape)} and the dimensions of the origin ("
                    f"image_origin; {len(self.image_origin)}) are different. [{self.describe_self()}]"
                )
            else:
                return False

        # Check that image_orientation has the correct number of dimensions, and is correctly formatted.
        if self.image_orientation is not None:
            if not np.all(np.equal(self.image_orientation.shape, len(data_shape))):
                if raise_error:
                    raise ValueError(
                        f"The orientation matrix should be square, with a dimension equal to the dimensions "
                        f"the image data itself ({len(data_shape)}. Found: {self.image_orientation.shape}."
                        f"[{self.describe_self()}]"
                    )
                else:
                    return False

        # Check that image orientation has a l2-norm of 1.0.
        if self.image_orientation is not None:
            l2_norm = np.around(np.linalg.norm(self.image_orientation, ord=2), decimals=3)
            if not l2_norm == 1.0:
                if raise_error:
                    raise ValueError(
                        f"The orientation matrix should be square with an l2-norm of 1.0. Found: {l2_norm}."
                        f"[{self.describe_self()}]"
                    )
                else:
                    return False

        # Check that spacing has the correct number of dimensions.
        if self.image_spacing is not None:
            if not len(self.image_spacing) == len(data_shape):
                if raise_error:
                    raise ValueError(
                        f"The dimensions of the image data itself ({len(data_shape)}) and the dimensions of the voxel "
                        f"spacing (image_spacing; {len(self.image_spacing)}) are different. [{self.describe_self()}]"
                    )
                else:
                    return False

        # Check that spacing contains strictly positive values.
        if self.image_spacing is not None:
            if np.any(np.array(self.image_spacing) <= 0.0):
                if raise_error:
                    raise ValueError(
                        f"Image spacing should be strictly positive. Found: {self.image_spacing}. "
                        f"[{self.describe_self()}]"
                    )
                else:
                    return False

    def _check_sample_name(self, raise_error: bool) -> bool:
        # Check that image file contains a sample name, if multiple sample names are present. To assess the filename,
        # we first strip the extension. Optionally we split the filename on the image name pattern, reducing the
        # filename into parts that should contain the sample name.
        if isinstance(self.sample_name, list) and len(self.sample_name) > 1:
            if self.get_sample_name_from_file() is None:
                if raise_error:
                    raise ValueError(
                        f"The file name of the image file {os.path.basename(self.file_path)} does not contain "
                        f"any of the expected patterns: {', '.join(self.sample_name)} [{self.describe_self()}]"
                    )
                else:
                    return False

        return True

    def _check_modality(self, raise_error: bool) -> bool:
        return True

    def get_sample_name_from_folder(self, sub_folder: None | str) -> None | str:
        """
        Aims to get a sample name candidate based on folder structure.
        """

        # Do not obtain sample name if it is already present.
        if isinstance(self.sample_name, str):
            return None

        # Do not obtain sample name if either file name or file path are missing
        if self.dir_path is None:
            return None

        dir_struct: list[str] = path_to_parts(self.dir_path)
        if sub_folder is None:
            dir_sample_name = dir_struct[-1]
        else:
            # Subtract sub_folder.
            for path_elem in reversed(path_to_parts(sub_folder)):
                if path_elem == "":
                    continue
                if dir_struct[-1] == path_elem:
                    dir_struct.pop()
                else:
                    raise ValueError(f"sub-folder structure `{path_elem}` not found on file path `{self.dir_path}`.")
            dir_sample_name = dir_struct[-1]

        if isinstance(self.sample_name, list) and len(self.sample_name) > 1:
            if dir_sample_name not in self.sample_name:
                return None

        return dir_sample_name

    def get_sample_name_from_file(self, must_succeed=False) -> None | str:
        allowed_file_extensions = supported_file_types(self.file_type)

        # Do not obtain sample name from the file name if a file name has already been set.
        if isinstance(self.sample_name, str):
            return None

        if self.file_name is None:
            return None

        # Select the most appropriate sample name.
        if isinstance(self.sample_name, list) and len(self.sample_name) > 1:
            file_name = bare_file_name(x=self.file_name, file_extension=allowed_file_extensions)
            if self.image_name is not None:
                image_id_name = self.image_name
                if not isinstance(image_id_name, list):
                    image_id_name = [image_id_name]

                # Find the id that is present in the filename.
                matching_image_id = None
                for current_image_id_name in image_id_name:
                    # Replace markers for the sample name (#).
                    current_image_id_name = current_image_id_name.replace("#", "*")
                    if fnmatch.fnmatch(file_name, current_image_id_name):
                        matching_image_id = current_image_id_name
                        break

                if matching_image_id is not None:
                    # Handle wildcards in the image id.
                    matching_image_id = matching_image_id.replace("?", "*")

                    # Split on wildcards.
                    matching_image_id = matching_image_id.split("*")

                    # Identify parts of the image id that are not empty after splitting.
                    matching_image_id = [x for x in matching_image_id if x != ""]

                    if len(matching_image_id) == 0:
                        file_name_parts = [file_name]
                    else:
                        # Find the parts of the file name that do not contain the image identifier.
                        blocked_start = file_name.find(matching_image_id[0])
                        blocked_end = file_name.find(matching_image_id[-1]) + len(matching_image_id[-1])
                        file_name_parts = [""]
                        if blocked_start > 0:
                            file_name_parts.append(file_name[0:blocked_start])

                        if blocked_end < len(file_name):
                            file_name_parts.append(file_name[blocked_end:len(file_name)])
                else:
                    file_name_parts = [file_name]

            else:
                file_name_parts = [file_name]

            # Check if any sample name is present.
            matching_name = None
            matching_frac = 0.0
            for current_file_name_part in file_name_parts:
                for current_sample_name in self.sample_name:
                    if current_sample_name in current_file_name_part:
                        # Prefer the most complete matches.
                        if len(current_sample_name) / len(current_file_name_part) > matching_frac:
                            matching_frac = len(current_sample_name) / len(current_file_name_part)
                            matching_name = current_sample_name

                            if matching_frac == 1.0:
                                return matching_name

            return matching_name

        elif must_succeed:
            return bare_file_name(x=self.file_name, file_extension=allowed_file_extensions)
        else:
            return None

    def _get_numeric_sequence_from_file(self) -> None | list[str]:
        allowed_file_extensions = supported_file_types(self.file_type)
        file_name = bare_file_name(x=self.file_name, file_extension=allowed_file_extensions)

        file_name_parts = [file_name]

        # Remove image name.
        if self.image_name is not None:
            image_id_name = self.image_name
            if not isinstance(image_id_name, list):
                image_id_name = [image_id_name]

            # Find the id that is present in the filename.
            matching_image_id = None
            for current_image_id_name in image_id_name:
                if current_image_id_name in file_name:
                    matching_image_id = current_image_id_name
                    break

            if matching_image_id is not None:
                # Handle wildcards in the image id.
                matching_image_id = matching_image_id.replace("?", "*")
                matching_image_id = matching_image_id.split("*")
                matching_image_id = [x for x in matching_image_id if x != ""]

                if len(matching_image_id) == 0:
                    file_name_parts = [file_name]
                else:
                    # Find the parts of the file name that do not contain the image identifier.
                    blocked_start = file_name.find(matching_image_id[0])
                    blocked_end = file_name.find(matching_image_id[-1]) + len(matching_image_id[-1])
                    file_name_parts = [""]
                    if blocked_start > 0:
                        file_name_parts.append(file_name[0:blocked_start])

                    if blocked_end < len(file_name):
                        file_name_parts.append(file_name[blocked_end:len(file_name)])
            else:
                file_name_parts = [file_name]

        # Remove empty strings (if any).
        file_name_parts = [
            current_file_name_part for current_file_name_part in file_name_parts
            if len(current_file_name_part) > 0
        ]

        if len(file_name_parts) == 0:
            return None

        # Remove sample name.
        if self.sample_name is not None:
            if isinstance(self.sample_name, list):
                raise TypeError(
                    f"The sample_name attribute cannot be a list for extracting numeric sequences. "
                    f"[{self.describe_self()}]"
                )

            file_name_parts = [
                current_file_name_part.split(sep=self.sample_name)
                for current_file_name_part in file_name_parts
            ]

            # Flatten list
            file_name_parts = list(itertools.chain.from_iterable(file_name_parts))

        # Remove empty strings (if any).
        file_name_parts = [
            current_file_name_part for current_file_name_part in file_name_parts
            if len(current_file_name_part) > 0
        ]

        if len(file_name_parts) == 0:
            return None

        # Isolate sequences of numeric values using regex. Then remove parts where no numeric value was found.
        file_name_parts = [re.findall(r'\d+', current_file_name_part) for current_file_name_part in file_name_parts]
        file_name_parts = [
            current_file_name_part for current_file_name_part in file_name_parts
            if len(current_file_name_part) > 0
        ]

        if len(file_name_parts) == 0:
            return None

        # Flatten list and return.
        return list(itertools.chain.from_iterable(file_name_parts))

    def complete(self, remove_metadata=False, force=False):
        self._complete_modality()
        self._complete_sample_name()
        self._complete_image_dimensions(force=force)
        self._complete_image_origin(force=force)
        self._complete_image_orientation(force=force)
        self._complete_image_spacing(force=force)

        # Check if the complete data passes verification.
        self.check(raise_error=True, remove_metadata=False)

        # Remove metadata. This allows file connections to be garbage collected.
        if remove_metadata:
            self.remove_metadata()

    def _complete_modality(self):
        # Set modality.
        if self.modality is None:
            self.modality = "generic"

    def _complete_sample_name(self):
        # Set sample name.
        if isinstance(self.sample_name, list):
            file_sample_name = self.get_sample_name_from_file()

            if file_sample_name is None and len(self.sample_name) == 1:
                self.sample_name = self.sample_name[0]

            elif file_sample_name is not None:
                self.sample_name = file_sample_name

            else:
                self.sample_name = None

    def _complete_image_origin(self, force=False, frame_id: None | int = None):
        raise NotImplementedError(
            f"DEV: There is (intentionally) no generic implementation of _complete_sample_origin. Please specify "
            f"implementation for subclasses."
        )

    def _complete_image_orientation(self, force=False, frame_id: None | int = None):
        raise NotImplementedError(
            f"DEV: There is (intentionally) no generic implementation of _complete_sample_orientation. Please specify "
            f"implementation for subclasses."
        )

    def _complete_image_spacing(self, force=False, frame_id: None | int = None):
        raise NotImplementedError(
            f"DEV: There is (intentionally) no generic implementation of _complete_sample_spacing. Please specify "
            f"implementation for subclasses."
        )

    def _complete_image_dimensions(self, force=False):
        raise NotImplementedError(
            f"DEV: There is (intentionally) no generic implementation of _complete_sample_dimensions. Please specify "
            f"implementation for subclasses."
        )

    def load_metadata(self, limited=False, include_image=False):
        raise NotImplementedError(
            f"DEV: There is (intentionally) no generic implementation of load_metadata. Please specify "
            f"implementation for subclasses."
        )

    def remove_metadata(self):
        self.image_metadata = None

    def load_data(self, **kwargs):
        """
        Loads data from file as numpy array, and attaches it to the current object by setting the image_data attribute.
        :param kwargs: Unused arguments.
        :return:
        """
        raise NotImplementedError(
            f"DEV: There is (intentionally) no generic implementation of load_data. Please specify "
            f"implementation for subclasses."
        )

    def stack_slices(self):
        """
        Stacks slices and sets the image_data attribute. Does nothing unless the image consists of stacks.
        :return:
        """
        pass

    def update_image_data(self):
        if self.image_data is None:
            return

        if not isinstance(self.image_data, np.ndarray):
            raise TypeError(
                f"The image_data argument expects None or a numpy ndarray. Found object with class "
                f"{type(self.image_data).name}. [{self.describe_self()}]"
            )

        # If the image is already 3D, we forgo adding dimensions.
        if len(self.image_data.shape) == 3:
            return

        if not 1 <= len(self.image_data.shape) <= 3:
            raise ValueError(
                f"MIRP supports image data up to 3 dimensions. The current numpy array has a dimension of "
                f"{len(self.image_data.shape)} ({self.image_data.shape}). [{self.describe_self()}]"
            )

        dims_to_add = 3 - len(self.image_data.shape)

        # Add new dimensions to image.
        if dims_to_add == 2:
            self.image_data = self.image_data[np.newaxis, np.newaxis, :]
        elif dims_to_add == 1:
            self.image_data = self.image_data[np.newaxis, :, :]

        # Update image dimensions attribute.
        if self.image_dimension is not None and not len(self.image_dimension) == 3:
            image_dimension = list(self.image_dimension)
            if dims_to_add == 2:
                image_dimension.insert(0, 1)
                image_dimension.insert(0, 1)
            elif dims_to_add == 1:
                image_dimension.insert(0, 1)

            self.image_dimension = tuple(image_dimension)

        # Update image origin attribute.
        if self.image_origin is not None and not len(self.image_origin) == 3:
            image_origin = list(self.image_origin)
            if dims_to_add == 2:
                image_origin.insert(0, 0)
                image_origin.insert(0, 0)
            elif dims_to_add == 1:
                image_origin.insert(0, 0)

            self.image_origin = tuple(image_origin)

        # Update image orientation attribute.
        if self.image_orientation is not None and not np.all(np.equal(self.image_orientation.shape, 3)):
            image_orientation = np.identity(3)
            if dims_to_add == 2:
                image_orientation[1:2, 1:2] = self.image_orientation
            elif dims_to_add == 1:
                image_orientation[1:3, 1:3] = self.image_orientation

            self.image_orientation = image_orientation

        # Update image spacing attribute.
        if self.image_spacing is not None and not len(self.image_spacing) == 3:
            image_spacing: list[float] = list(self.image_spacing)
            if dims_to_add == 2:
                image_spacing.insert(0, image_spacing[0])
                image_spacing.insert(0, 1.0)
            if dims_to_add == 1:
                image_spacing.insert(0, 1.0)

            self.image_spacing = tuple(image_spacing)

    def to_object(self, **kwargs) -> GenericImage:

        self.load_data(**kwargs)
        self.complete()
        self.stack_slices()
        self.update_image_data()
        self.set_object_metadata()

        return GenericImage(
            sample_name=self.sample_name,
            image_modality=self.modality,
            image_data=self.image_data,
            image_spacing=self.image_spacing,
            image_origin=self.image_origin,
            image_orientation=self.image_orientation,
            image_dimensions=self.image_dimension,
            metadata=self.object_metadata
        )

    def set_object_metadata(self):
        """
        Updates the object metadata that is passed to native image and mask classes in to_object.
        """
        attributes = []
        # Add file name to object metadata.
        if self.file_name is not None:
            attributes += [("file_name", self.file_name)]

        # Add directory to object metadata.
        if self.dir_path is not None:
            attributes += [("dir_path", self.dir_path)]

        # Add file type to object metadata.
        if self.file_type is not None:
            attributes += [("file_type", self.file_type)]

        self.object_metadata.update(dict(attributes))

    def export_metadata(self, **kwargs) -> None | dict[str, Any]:
        # Get general attributes.
        parent_attributes = self._get_export_attributes()

        # Add metadata.
        metadata = []
        if self.image_spacing is not None:
            metadata += [
                ("spacing_z", self.image_spacing[0]),
                ("spacing_y", self.image_spacing[1]),
                ("spacing_x", self.image_spacing[2])
            ]
        parent_attributes.update(dict(metadata))

        return parent_attributes

    def _get_export_attributes(self) -> dict[str, Any]:
        attributes = []
        if self.sample_name is not None:
            attributes += [("sample_name", self.sample_name)]
        if self.modality is not None:
            attributes += [("modality", self.modality)]
        if self.dir_path is not None:
            attributes += [("dir_path", self.dir_path)]
        if self.file_name is not None:
            attributes += [("file_name", self.file_name)]

        return dict(attributes)

    def describe_self(self):
        return f"{self._get_export_attributes()}"

    def check_associated_masks(self):
        if self.associated_masks is None:
            return

        for mask in self.associated_masks:
            self._check_associated_mask_image_data(mask=mask)

    def _check_associated_mask_image_data(self, mask: Self):
        """Check whether image and mask plausibly share the same frame of reference."""

        problem_list = []
        # Mismatch in grid dimension
        if isinstance(self.image_dimension, tuple) and isinstance(mask.image_dimension, tuple):
            if not np.array_equal(self.image_dimension, mask.image_dimension):
                problem_list += [
                    f"different dimensions: \n\t\timage: {self.image_dimension}\n\t\tmask: {mask.image_dimension}"
                ]

        # Mismatch in origin
        if isinstance(self.image_origin, tuple) and isinstance(mask.image_origin, tuple):
            if not np.allclose(self.image_origin, mask.image_origin):
                problem_list += [
                    f"different origin: \n\t\timage: {self.image_origin}\n\t\tmask: {mask.image_origin}"
                ]

        # Mismatch in spacing
        if isinstance(self.image_spacing, tuple) and isinstance(mask.image_spacing, tuple):
            if not np.allclose(self.image_spacing, mask.image_spacing):
                problem_list += [
                    f"different spacing: \n\t\timage: {self.image_spacing}\n\t\tmask: {mask.image_spacing}"
                ]

        # Mismatch in orientation
        if isinstance(self.image_orientation, np.ndarray) and isinstance(mask.image_orientation, np.ndarray):
            if not np.allclose(self.image_orientation, mask.image_orientation):
                problem_list += [
                    f"different orientation: \n\t\timage: {np.ravel(self.image_orientation)}\n\t\tmask: "
                    f"{np.ravel(mask.image_orientation)}"
                ]

        if len(problem_list) > 0:
            warnings.warn(
                f"Image {self.describe_self()} and mask {mask.describe_self()} may not have the same frame of "
                f"reference. Please check if segmentation masks are placed correctly:\n\t" + "\n\t".join(problem_list),
                UserWarning
            )


class MaskFile(ImageFile):

    def __init__(
            self,
            roi_name: None | str | list[str] | dict[str, str] = None,
            **kwargs
    ):
        super().__init__(**kwargs)

        # Curate input.
        if roi_name is None or isinstance(roi_name, str) or \
            (isinstance(roi_name, list) and all(isinstance(x, str) for x in roi_name)) or \
                (isinstance(roi_name, dict) and all(isinstance(x, str) for x in roi_name.values()) and all(
                    isinstance(x, str) for x in roi_name.keys())):
            self.roi_name = roi_name
        else:
            raise TypeError(
                f"ROI names are expected to be a string, a list of strings, or a dictionary of strings. Found:"
                f" {roi_name} with type {type(roi_name)}. [{self.describe_self()}]")

    def is_stackable(self, stack_images: str):
        raise NotImplementedError(
            f"DEV: There is (intentionally) no generic implementation of is_stackable. Please specify "
            f"implementation for subclasses."
        )

    def _complete_modality(self):
        # Set modality.
        if self.modality is None:
            self.modality = "generic_mask"

    def _complete_image_origin(self, force=False, frame_id=None):
        raise NotImplementedError(
            f"DEV: There is (intentionally) no generic implementation of _complete_image_origin. Please specify "
            f"implementation for subclasses."
        )

    def _complete_image_orientation(self, force=False, frame_id=None):
        raise NotImplementedError(
            f"DEV: There is (intentionally) no generic implementation of _complete_image_orientation. Please specify "
            f"implementation for subclasses."
        )

    def _complete_image_spacing(self, force=False, frame_id=None):
        raise NotImplementedError(
            f"DEV: There is (intentionally) no generic implementation of _complete_image_spacing. Please specify "
            f"implementation for subclasses."
        )

    def _complete_image_dimensions(self, force=False):
        raise NotImplementedError(
            f"DEV: There is (intentionally) no generic implementation of _complete_image_dimensions. Please specify "
            f"implementation for subclasses."
        )

    def load_metadata(self, **kwargs):
        raise NotImplementedError(
            f"DEV: There is (intentionally) no generic implementation of load_metadata. Please specify "
            f"implementation for subclasses."
        )

    def load_data(self, **kwargs):
        raise NotImplementedError(
            f"DEV: There is (intentionally) no generic implementation of load_data. Please specify "
            f"implementation for subclasses."
        )

    def set_modality(self, modality: None | str):
        from mirp._data_import.utilities import supported_mask_modalities
        if modality is None:
            return

        if not isinstance(modality, str):
            raise ValueError(
                f"modality is expected to be a character string. Found: {modality}. [{self.describe_self()}]"
            )

        if modality == "generic_mask":
            raise ValueError(f"modality cannot be 'generic_mask'. [{self.describe_self()}]")

        modality = supported_mask_modalities(modality)
        if self.modality is None or self.modality == "generic_mask":
            self.modality = modality[0]

    def create(self):
        # Import locally to avoid potential circular references.
        from mirp._data_import.dicom_file import MaskDicomFile
        from mirp._data_import.itk_file import MaskITKFile
        from mirp._data_import.numpy_file import MaskNumpyFile

        file_extensions = supported_file_types(file_type=self.file_type)

        if any(self.file_path.lower().endswith(ii) for ii in file_extensions) and\
                any(self.file_path.lower().endswith(ii) for ii in supported_file_types("dicom")):

            file_class = MaskDicomFile
            file_type = "dicom"

        elif any(self.file_path.lower().endswith(ii) for ii in file_extensions) and\
                any(self.file_path.lower().endswith(ii) for ii in supported_file_types("itk")):
            if any(self.file_path.lower().endswith(ii) for ii in supported_file_types("nifti")):
                file_type = "nifti"
            elif any(self.file_path.lower().endswith(ii) for ii in supported_file_types("nrrd")):
                file_type = "nrrd"
            else:
                raise ValueError(f"DEV: specify file_type")

            file_class = MaskITKFile

        elif any(self.file_path.lower().endswith(ii) for ii in file_extensions) and\
                any(self.file_path.lower().endswith(ii) for ii in supported_file_types("numpy")):

            file_class = MaskNumpyFile
            file_type = "numpy"

        else:
            raise NotImplementedError(
                f"The provided mask type is not implemented: {self.file_type}. [{self.describe_self()}]"
            )

        image_file = file_class(
            file_path=self.file_path,
            dir_path=self.dir_path,
            sample_name=self.sample_name,
            file_name=self.file_name,
            image_name=self.image_name,
            image_modality=self.modality,
            image_file_type=file_type,
            image_data=self.image_data,
            image_origin=self.image_origin,
            image_orientation=self.image_orientation,
            image_spacing=self.image_spacing,
            image_dimensions=self.image_dimension,
            roi_name=self.roi_name
        ).create()

        return image_file

    def check_mask(self, raise_error=True):
        if self.image_data is None:
            raise TypeError(f"DEV: the image_data attribute has not been set. [{self.describe_self()}]")

        if np.issubdtype(self.image_data.dtype, bool):

            if not np.any(self.image_data):
                if raise_error:
                    warnings.warn(
                        f"No regions of interest were formed. "
                        f"The mask object only contains background values (False). "
                        f"No voxels were found with True values to identify segmentation masks. "
                        f"[{self.describe_self()}]",
                        UserWarning
                    )
            return True

        if np.issubdtype(self.image_data.dtype, np.integer):
            if np.any(self.image_data < 0):
                if raise_error:
                    raise ValueError(
                        f"Labels in a mask should be 0 or positive integers. Negative values were found. "
                        f"Note that 0 is interpreted as background. [{self.describe_self()}]"
                    )
                return False

            if len(np.unique(self.image_data)) > 10:
                if raise_error:
                    warnings.warn(
                        f"More than 10 labels were found. Please check that this is correct. "
                        f"[{self.describe_self()}]",
                        UserWarning
                    )

            if np.all(self.image_data == 0):
                if raise_error:
                    warnings.warn(
                        f"No regions of interest were formed. The mask object only contains "
                        f"background values (0). No voxels were found with positive integers to identify segmentation "
                        f"masks. [{self.describe_self()}]",
                        UserWarning
                    )

        else:
            raise TypeError(f"DEV: the image_data contains non-integer data ({self.image_data.dtype})")

        return True

    def to_object(self, **kwargs) -> None | list[BaseMask]:

        self.load_data()
        self.complete()
        self.stack_slices()
        self.update_image_data()
        self.set_object_metadata()
        self.check_mask(raise_error=True)

        mask_list = []
        if np.issubdtype(self.image_data.dtype, bool):
            if not np.any(self.image_data):
                return None

            if isinstance(self.roi_name, str):
                roi_name = self.roi_name
            elif isinstance(self.roi_name, list):
                if len(self.roi_name) == 1:
                    roi_name = self.roi_name[0]
                else:
                    warnings.warn(
                        f"The name of the region of interest could not be unambiguously determined. "
                        f"One of the following should be used, but it is not clear which: {self.roi_name}. "
                        f"[{self.describe_self()}]",
                        UserWarning
                    )
                    roi_name = "region_1"
            elif isinstance(self.roi_name, dict):
                if len(self.roi_name) == 1:
                    roi_name = list(self.roi_name.values())[0]
                else:
                    warnings.warn(
                        f"The name of the region of interest could not be unambiguously determined. "
                        f"One of the following should be used, but it is not clear which: {self.roi_name}. "
                        f"[{self.describe_self()}]",
                        UserWarning
                    )
                    roi_name = "region_1"
            else:
                roi_name = "region_1"

            mask_list += [
                BaseMask(
                    roi_name=roi_name,
                    sample_name=self.sample_name,
                    image_modality=self.modality,
                    image_data=self.image_data,
                    image_spacing=self.image_spacing,
                    image_origin=self.image_origin,
                    image_orientation=self.image_orientation,
                    image_dimensions=self.image_dimension,
                    metadata=self.object_metadata
                )
            ]

        else:

            # Identify available labels that are non-background.
            labels = list(set(np.unique(self.image_data)) - {0})

            if len(labels) == 0:
                return None

            # Check which labels should be kept.
            filtered_labels = []
            for current_label in labels:
                if self.roi_name is None:
                    filtered_labels += [current_label]
                elif isinstance(self.roi_name, str):
                    if self.roi_name == str(current_label) or self.roi_name == "region_" + str(current_label):
                        filtered_labels += [current_label]
                elif isinstance(self.roi_name, list):
                    if str(current_label) in self.roi_name or "region_" + str(current_label) in self.roi_name:
                        filtered_labels += [current_label]
                elif isinstance(self.roi_name, dict):
                    if str(current_label) in self.roi_name.keys() or "region_" + str(current_label) in self.roi_name.keys():
                        filtered_labels += [current_label]

            if len(filtered_labels) == 0:
                warnings.warn(
                    f"No regions of interest were formed. "
                    f"The available labels ({', '.join([str(x) for x in labels])}) likely did not match "
                    f"any of the expected labels ({self.roi_name}). "
                    f"[{self.describe_self()}]",
                    UserWarning
                )

                return None

            for current_label in filtered_labels:

                # Find roi name to use for the current object.
                if isinstance(self.roi_name, str):
                    roi_name = self.roi_name
                elif isinstance(self.roi_name, list):
                    roi_name = [x for x in self.roi_name if str(current_label) == x or "region_" + str(current_label) == x][0]
                elif isinstance(self.roi_name, dict):
                    roi_name = self.roi_name.get(str(current_label))
                    if roi_name is None:
                        roi_name = self.roi_name.get("region_" + str(current_label))
                else:
                    roi_name = "region_" + str(current_label)

                mask_list += [
                    BaseMask(
                        roi_name=roi_name,
                        sample_name=self.sample_name,
                        image_modality=self.modality,
                        image_data=self.image_data == current_label,
                        image_spacing=self.image_spacing,
                        image_origin=self.image_origin,
                        image_orientation=self.image_orientation,
                        image_dimensions=self.image_dimension
                    )
                ]

        return mask_list

    def export_metadata(self) -> dict[str, Any]:
        return super().export_metadata()

    def export_roi_labels(self) -> dict[str, Any]:

        if self.image_data is None:
            self.load_data()
            self.complete()
            self.stack_slices()

        # Identify available labels that are non-background.
        labels = list(set(np.unique(self.image_data)) - {0})
        n_labels = max([1, len(labels)])

        if len(labels) == 0:
            labels = [None]

        # Get general attributes.
        parent_attributes = self._get_export_attributes()

        # Add roi labels as attribute.
        attributes = [("roi_label", labels)]
        parent_attributes.update(dict(attributes))

        return parent_attributes


class MaskFullImage(MaskFile):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.roi_name = "full_image_mask"

    def to_object(
            self,
            image: None | ImageFile,
            **kwargs
    ) -> None | list[BaseMask]:
        if image is None:
            raise TypeError(
                f"Creation of a full image mask requires that the corresponding image is set. "
                f"No image was provided. [{self.describe_self()}]"
            )
        else:
            image.complete()

        self._complete_modality()

        return [BaseMask(
            roi_name=self.roi_name,
            sample_name=image.sample_name,
            image_modality=self.modality,
            image_data=np.ones(image.image_dimension, dtype=bool),
            image_spacing=image.image_spacing,
            image_origin=image.image_origin,
            image_orientation=image.image_orientation,
            image_dimensions=image.image_dimension,
            metadata=self.object_metadata
        )]



import numpy as np
import copy

from mirp._images.generic_image import GenericImage
from mirp._images.transformed_image import SeparableWaveletTransformedImage
from mirp.settings.generic import SettingsClass
from mirp._imagefilters.generic import GenericFilter
from mirp._imagefilters.utilities import pool_voxel_grids, SeparableFilterSet


class SeparableWaveletFilter(GenericFilter):

    def __init__(self, image: GenericImage, settings: SettingsClass, name: str):

        super().__init__(image=image, settings=settings, name=name)

        self.ibsi_compliant = True
        self.ibsi_id = "25BO"

        # Set wavelet family
        self.wavelet_family: str | list[str] = settings.img_transform.separable_wavelet_families

        # Wavelet decomposition level
        self.decomposition_level: int | list[int] = settings.img_transform.separable_wavelet_decomposition_level

        # Set the filter set for separable wavelets.
        self.filter_configuration: str | list[str] = settings.img_transform.separable_wavelet_filter_set

        # Set rotational invariance
        self.rotational_invariance = settings.img_transform.separable_wavelet_rotation_invariance

        # Which pooling method is used.
        self.pooling_method = settings.img_transform.separable_wavelet_pooling_method

        # Wavelet cascade type
        self.stationary_wavelet = settings.img_transform.separable_wavelet_stationary

        # Set boundary condition
        self.mode = settings.img_transform.separable_wavelet_boundary_condition

    def generate_object(self):
        # Generator for transformation objects.
        wavelet_family = copy.deepcopy(self.wavelet_family)
        if not isinstance(wavelet_family, list):
            wavelet_family = [wavelet_family]

        filter_configuration = copy.deepcopy(self.filter_configuration)
        if not isinstance(filter_configuration, list):
            filter_configuration = [filter_configuration]

        decomposition_level = copy.deepcopy(self.decomposition_level)
        if not isinstance(decomposition_level, list):
            decomposition_level = [decomposition_level]

        # Iterate over options to yield filter objects with specific settings. A copy of the parent object is made to
        # avoid updating by reference.
        for current_wavelet_family in wavelet_family:
            for current_filter_configuration in filter_configuration:
                for current_decomposition_level in decomposition_level:
                    filter_object = copy.deepcopy(self)
                    filter_object.wavelet_family = current_wavelet_family
                    filter_object.filter_configuration = current_filter_configuration
                    filter_object.decomposition_level = current_decomposition_level

                    yield filter_object

    def transform(self, image: GenericImage) -> SeparableWaveletTransformedImage:
        # Create placeholder separable wavelet response map.
        response_map = SeparableWaveletTransformedImage(
            image_data=None,
            wavelet_family=self.wavelet_family,
            decomposition_level=self.decomposition_level,
            filter_kernel_set=self.filter_configuration,
            stationary_wavelet=self.stationary_wavelet,
            rotation_invariance=self.rotational_invariance,
            pooling_method=self.pooling_method,
            boundary_condition=self.mode,
            riesz_order=None,
            riesz_steering=None,
            riesz_sigma_parameter=None,
            template=image
        )
        response_map.ibsi_compliant = self.ibsi_compliant and image.ibsi_compliant

        if image.is_empty():
            return response_map

        # Initialise voxel grid.
        response_voxel_grid = None

        # Get filter list.
        filter_set_list: list[SeparableFilterSet] = self.get_filter_set().permute_filters(
            rotational_invariance=self.rotational_invariance,
            require_pre_filter=True
        )

        for ii, filter_set in enumerate(filter_set_list):

            # Extract the voxel grid as starting point.
            pooled_voxel_grid = image.get_voxel_grid()

            for decomposition_level in np.arange(1, self.decomposition_level + 1):

                # Determine whether the pre-filter should be applied. This is the case for decomposition levels
                # smaller than self.decomposition_level.
                use_pre_filter = decomposition_level < self.decomposition_level

                # Convolve and compute the response map.
                pooled_voxel_grid = filter_set.convolve(
                    voxel_grid=pooled_voxel_grid,
                    mode=self.mode,
                    use_pre_filter=use_pre_filter
                )

                if use_pre_filter:
                    # Decompose the filter set for the next level.
                    filter_set.decompose_filter()

            # Pool grids.
            response_voxel_grid = pool_voxel_grids(
                x1=response_voxel_grid,
                x2=pooled_voxel_grid,
                pooling_method=self.pooling_method
            )

            # Remove pooled_voxel_grid to explicitly release memory when collecting garbage.
            del pooled_voxel_grid

        if self.pooling_method == "mean":
            # Perform final pooling step for mean pooling.
            response_voxel_grid = np.divide(response_voxel_grid, len(filter_set_list))

        # Store the voxel grid in the ImageObject.
        response_map.set_voxel_grid(voxel_grid=response_voxel_grid)

        return response_map

    def get_filter_set(self):
        import pywt
        from copy import deepcopy

        # Deparse convolution kernels to a list
        kernel_list = [self.filter_configuration[ii:ii + 1] for ii in range(0, len(self.filter_configuration), 1)]

        # Declare filter kernels
        filter_x, filter_y, filter_z = None, None, None
        pre_filter_x, pre_filter_y, pre_filter_z = None, None, None

        # Define the pre-filter kernel for decomposition.
        pre_filter_kernel = np.array(pywt.Wavelet(self.wavelet_family).dec_lo)

        for ii, kernel in enumerate(kernel_list):
            if kernel.lower() == "l":
                wavelet_kernel = np.array(pywt.Wavelet(self.wavelet_family).dec_lo)
            elif kernel.lower() == "h":
                wavelet_kernel = np.array(pywt.Wavelet(self.wavelet_family).dec_hi)
            else:
                raise ValueError(f"{kernel} was not recognised as the component of a separable wavelet filter. It "
                                 f"should be L or H.")

            # Assign filter to variable.
            if ii == 0:
                filter_x = wavelet_kernel
                pre_filter_x = deepcopy(pre_filter_kernel)
            elif ii == 1:
                filter_y = wavelet_kernel
                pre_filter_y = deepcopy(pre_filter_kernel)
            elif ii == 2:
                filter_z = wavelet_kernel
                pre_filter_z = deepcopy(pre_filter_kernel)

        # Create FilterSet object
        return SeparableFilterSet(filter_x=filter_x,
                                  filter_y=filter_y,
                                  filter_z=filter_z,
                                  pre_filter_x=pre_filter_x,
                                  pre_filter_y=pre_filter_y,
                                  pre_filter_z=pre_filter_z)



import warnings

def parse_parallel_backend(backend: None | str, num_cpus: None | int, ray_allowed: bool = True) -> str:
    from mirp.utilities.parallel_ray import ray_is_available, ray_is_initialized
    from mirp.utilities.parallel_joblib import joblib_is_available

    if num_cpus is None or num_cpus <= 1 or backend == "none":
        return "none"

    if backend == "ray":
        if not ray_allowed:
            warnings.warn(
                f"The ray module cannot be used for parallel processing within the current context. Sequential processing is used.",
                UserWarning
            )
            return "none"

        if ray_is_available():
            return "ray"
        warnings.warn(
            f"Parallel processing requires that either ray module is installed. "
            f"However, the ray module could not be imported. Sequential processing is used.",
            UserWarning
        )
    elif backend == "joblib":
        if joblib_is_available():
            return "joblib"
        warnings.warn(
            f"Parallel processing requires that either joblib module is installed. "
            f"However, the joblib module could not be imported. Sequential processing is used.",
            UserWarning
        )
    else:
        raise ValueError(f"backend is expected to be one of 'none', 'ray' or 'joblib'. Found: {backend}")

    if backend is None:
        if ray_is_available():
            return "ray"
        if joblib_is_available():
            return "joblib"
        warnings.warn(
            f"Parallel processing requires that either joblib or ray modules are installed. "
            f"These modules could not be imported. Sequential processing is used.",
            UserWarning
        )

    return "none"


def start_parallel_cluster(backend: str, num_cpus: int):
    if backend == "ray":
        from mirp.utilities.parallel_ray import ray_is_initialized, ray_init
        if not cluster_exists(backend=backend):
            ray_init(num_cpus=num_cpus)

    return


def cluster_exists(backend: str) -> bool:
    if backend == "ray":
        from mirp.utilities.parallel_ray import ray_is_initialized
        return ray_is_initialized()

    return False


def shutdown_cluster(backend: str):
    if backend == "ray":
        from mirp.utilities.parallel_ray import ray_shutdown
        ray_shutdown()

def message_parallel_process(backend: str, num_cpus: None | int) -> str:
    if backend == "none":
        return "Jobs are processed sequentially."
    elif backend == "ray":
        return f"Jobs are processed in parallel using the ray backend with {num_cpus} workers."
    elif backend == "joblib":
        return f"Jobs are processed in parallel using the joblib backend with {num_cpus} workers."
    else:
        raise ValueError(f"backend is expected to be one of 'none', 'ray' or 'joblib'. Found: {backend}")


import copy

import pandas as pd
import numpy as np
import sys

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self
from typing import Generator

from mirp._images.generic_image import GenericImage
from mirp._masks.base_mask import BaseMask


class Matrix(object):
    def __init__(
            self,
            spatial_method: str | None,
            matrix: pd.DataFrame | None = None,
            slice_id: int | None = None,
            n_voxels: int | None = None
    ):
        # Placeholder for the actual matrix.
        self.matrix = matrix

        # Spatial method for aggregating and merging matrices.
        self.spatial_method = spatial_method

        # Slice ID that is required for merging matrices within a slice.
        self.slice_id = slice_id

        # Number of voxels represented in the matrix.
        self.n_voxels = n_voxels

    def copy(self):
        return copy.deepcopy(self)

    def is_empty(self):
        return self.matrix is None or len(self.matrix) == 0

    def compute(
            self,
            data: pd.DataFrame | None = None,
            image: GenericImage | None = None,
            mask: BaseMask | None = None,
            image_dimension: tuple[int, int, int] | None = None,
            **kwargs
    ):
        raise NotImplementedError("Implement in subclasses.")

    def set_values_from_matrix(self, **kwargs):
        raise NotImplementedError("Implement in subclasses.")

    def generate(self, prototype, n_slices: int, **kwargs) -> Generator[Self, None, None]:
        if self.spatial_method in ["2d", "2.5d"]:
            # Extract texture matrices from 2D slices.
            for slice_id in np.arange(0, n_slices):
                yield prototype(
                    spatial_method=self.spatial_method,
                    slice_id=slice_id,
                    **kwargs
                )

        elif self.spatial_method == "3d":
            # Extract texture matrix for the whole 3D volume.
            yield prototype(
                spatial_method=self.spatial_method,
                **kwargs
            )

        else:
            raise ValueError(
                f"One of  \"2d\", \"2.5d\" or \"3d\" is expected as spatial method. Found: {self.spatial_method}."
            )

    def merge(
            self,
            matrix_list: list[Self],
            prototype,
            **kwargs
    ) -> list[Self]:

        if self.spatial_method == "2d":
            # Average features over slices: maintain original 2D texture matrices.
            return [matrix for matrix in matrix_list if not matrix.is_empty()]

        elif self.spatial_method in ["2.5d", "3d"]:
            # Merge 2D or 3D texture matrices into a single matrix.

            # Remove all empty matrices.
            updated_matrix_list = [matrix for matrix in matrix_list if not matrix.is_empty()]
            if len(updated_matrix_list) == 0:
                return []

            collected_matrices = [matrix.matrix for matrix in updated_matrix_list]
            n_voxels = np.sum([matrix.n_voxels for matrix in updated_matrix_list])

            # Merge collected matrices.
            merged_matrix = pd.concat(collected_matrices, axis=0)
            merged_matrix = merged_matrix.groupby(by=self._get_grouping_columns()).sum().reset_index()

            return [prototype(
                spatial_method=self.spatial_method,
                matrix=merged_matrix,
                n_voxels=n_voxels,
                **kwargs
            )]

    @staticmethod
    def _get_grouping_columns() -> list[str, ...]:
        raise NotImplementedError("Implement in subclasses.")

    @staticmethod
    def coord_to_index(x, y, z, dims):
        # Translate coordinates to indices
        index = x + y * dims[2] + z * dims[2] * dims[1]

        # Mark invalid transitions
        index[np.logical_or(x < 0, x >= dims[2])] = -99999
        index[np.logical_or(y < 0, y >= dims[1])] = -99999
        index[np.logical_or(z < 0, z >= dims[0])] = -99999

        return index

    @staticmethod
    def _generate_neighbour_direction(
            d: float = 1.8,
            metric: str = "euclidian",
            keep_centre: bool = False,
            complete: bool = False,
            dim3: bool = True
    ) -> Generator[tuple[int, int, int], None, None]:
        from mirp._features.utilities import rep

        # Base transition vector
        trans = np.arange(start=-np.ceil(d), stop=np.ceil(d) + 1)
        n = np.size(trans)

        # Build transition array [z,y,x]
        nbrs = np.array([
            rep(x=trans, each=1, times=n * n),
            rep(x=trans, each=n, times=n),
            rep(x=trans, each=n * n, times=1)
        ], dtype=np.int32)

        # Initiate maintenance index
        index = np.zeros(np.shape(nbrs)[1], dtype=bool)

        # Remove neighbours more than distance d from the center.
        if metric.lower() in ["manhattan", "l1", "l_1"]:
            # Manhattan distance
            index = np.logical_or(index, np.sum(np.abs(nbrs), axis=0) <= d)
        elif metric.lower() in ["euclidian", "l2", "l_2"]:
            # Euclidian distance
            index = np.logical_or(index, np.sqrt(np.sum(np.multiply(nbrs, nbrs), axis=0)) <= d)
        elif metric in ["chebyshev", "linf", "l_inf"]:
            # Chebyshev distance
            index = np.logical_or(index, np.max(np.abs(nbrs), axis=0) <= d)
        else:
            raise ValueError(f"Did not recognize distance metric: {metric}")

        # Check if centre voxel [0,0,0] should be maintained; False indicates removal
        if not keep_centre:
            index = np.logical_and(index, (np.sum(np.abs(nbrs), axis=0)) > 0)

        # Check if a complete neighbourhood should be returned
        # False indicates that only half of the vectors are returned
        if not complete:
            index[np.arange(start=0, stop=len(index)//2 + 1)] = False

        # Check if neighbourhood should be 3D or 2D
        if not dim3:
            index[nbrs[0, :] != 0] = False

        for ii, flag in enumerate(index):
            if flag:
                yield tuple(nbrs[:, ii].flatten())

    @staticmethod
    def _lookup_intensity(x, index, replace_invalid="nan"):
        # Initialise placeholder
        read_x = np.zeros(np.shape(x))

        # Read variables for valid indices
        read_x[index >= 0] = x[index[index >= 0]]

        if replace_invalid == "nan":
            # Set variables for invalid indices to nan
            read_x[index < 0] = np.nan

            # Set variables for invalid initial indices to nan
            read_x[np.isnan(x)] = np.nan

        return read_x

    def _spatial_method_error(self):
        raise ValueError(
            f"One of  \"2d\", \"2.5d\" or \"3d\" is expected as spatial method. Found: {self.spatial_method}."
        )


class DirectionalMatrix(Matrix):

    def __init__(
            self,
            direction_id: int | None = None,
            direction: tuple[int, int, int] | None = None,
            **kwargs
    ):
        super().__init__(**kwargs)
        # Set the neighbourhood directions and its index.
        self.direction_id = direction_id
        self.direction = direction

    def generate(self, prototype, n_slices: int, **kwargs) -> Generator[Self, None, None]:
        if self.spatial_method in ["2d_average", "2d_slice_merge", "2.5d_direction_merge", "2.5d_volume_merge"]:
            # Extract directional texture matrices from 2D slices.
            for slice_id in np.arange(0, n_slices):
                for direction_id, direction in enumerate(self._generate_neighbour_direction(
                    d=1,
                    metric="chebyshev",
                    dim3=False
                )):
                    yield prototype(
                        spatial_method=self.spatial_method,
                        slice_id=slice_id,
                        direction=direction,
                        direction_id=direction_id,
                        **kwargs
                    )

        elif self.spatial_method in ["3d_average", "3d_volume_merge"]:
            # Extract texture matrices for the whole 3D volume.
            for direction_id, direction in enumerate(self._generate_neighbour_direction(
                    d=1,
                    metric="chebyshev",
                    dim3=True
            )):
                yield prototype(
                    spatial_method=self.spatial_method,
                    slice_id=None,
                    direction=direction,
                    direction_id=direction_id,
                    **kwargs
                )

        else:
            self._spatial_method_error()

    def merge(
            self,
            matrix_list: list[Self],
            prototype,
            **kwargs
    ) -> list[Self]:

        # Remove empty matrices.
        matrix_list = [matrix for matrix in matrix_list if not matrix.is_empty()]
        if len(matrix_list) == 0:
            return []

        if self.spatial_method in ["2d_average", "3d_average"]:
            # Average features over directions: maintain original directional texture matrices.
            return matrix_list

        elif self.spatial_method == "2d_slice_merge":
            # Merge directional matrices within each slice.
            out_matrix_list = []

            # Identify slice ids and iterate over unique slice identifiers.
            slice_ids = [matrix.slice_id for matrix in matrix_list]
            for slice_id in np.unique(slice_ids):
                slice_matrix_list = [matrix for matrix in matrix_list if matrix.slice_id == slice_id]
                collected_matrices = [matrix.matrix for matrix in slice_matrix_list]
                n_voxels = np.sum([matrix.n_voxels for matrix in slice_matrix_list])

                merged_matrix = pd.concat(collected_matrices, axis=0)
                merged_matrix = merged_matrix.groupby(by=self._get_grouping_columns()).sum().reset_index()

                out_matrix_list += [prototype(
                    spatial_method=self.spatial_method,
                    matrix=merged_matrix,
                    n_voxels=n_voxels,
                    **kwargs
                )]

            return out_matrix_list

        elif self.spatial_method == "2.5d_direction_merge":
            # Merge directional matrices for each direction.
            out_matrix_list = []

            # Identify direction ids and iterate over unique direction identifiers.
            direction_ids = [matrix.direction_id for matrix in matrix_list]
            for direction_id in np.unique(direction_ids):
                direction_matrix_list = [matrix for matrix in matrix_list if matrix.direction_id == direction_id]
                collected_matrices = [matrix.matrix for matrix in direction_matrix_list]
                n_voxels = np.sum([matrix.n_voxels for matrix in direction_matrix_list])

                merged_matrix = pd.concat(collected_matrices, axis=0)
                merged_matrix = merged_matrix.groupby(by=self._get_grouping_columns()).sum().reset_index()

                out_matrix_list += [prototype(
                    spatial_method=self.spatial_method,
                    matrix=merged_matrix,
                    n_voxels=n_voxels,
                    **kwargs
                )]

            return out_matrix_list

        elif self.spatial_method in ["2.5d_volume_merge", "3d_volume_merge"]:
            # Merge 2D or 3D texture matrices into a single matrix.

            # Remove all empty matrices.
            updated_matrix_list = [matrix for matrix in matrix_list if not matrix.is_empty()]
            if len(updated_matrix_list) == 0:
                return []

            collected_matrices = [matrix.matrix for matrix in updated_matrix_list]
            n_voxels = np.sum([matrix.n_voxels for matrix in updated_matrix_list])

            # Merge collected matrices.
            merged_matrix = pd.concat(collected_matrices, axis=0)
            merged_matrix = merged_matrix.groupby(by=self._get_grouping_columns()).sum().reset_index()

            return [prototype(
                spatial_method=self.spatial_method,
                matrix=merged_matrix,
                n_voxels=n_voxels,
                **kwargs
            )]

        else:
            self._spatial_method_error()

    def _spatial_method_error(self):
        raise ValueError(
            f"One of  \"2d_average\", \"2d_slice_merge\", \"2.5d_direction_merge\", \"2.5d_volume_merge\", "
            f"\"3d_average\" or \"3d_volume_merge\" is expected as spatial method. Found: {self.spatial_method}."
        )



import os
import os.path
import copy

from itertools import chain
from mirp._data_import.utilities import supported_file_types, dir_structure_contains_directory, match_file_name, \
    isolate_sample_name, lookup_modality
from mirp._data_import.generic_file import ImageFile, MaskFile
from mirp._data_import.generic_file_stack import ImageFileStack, MaskFileStack


class ImageDirectory:

    def __init__(
            self,
            directory,
            sample_name: None | str | list[str] = None,
            image_name: None | str = None,
            image_sub_folder: None | str = None,
            image_modality: None | str = None,
            image_file_type: None | str = None,
            stack_images: str = "auto",
            **kwargs
    ):

        if sample_name is not None and not isinstance(sample_name, list):
            sample_name = [sample_name]

        self.image_directory = os.path.normpath(directory)
        self.sample_name: None | list[str] = sample_name
        self.image_name = image_name
        self.sub_folder = os.path.normpath(image_sub_folder) if image_sub_folder is not None else None

        if isinstance(image_modality, str):
            image_modality = image_modality.lower()
        self.modality: None | str = image_modality

        if isinstance(image_file_type, str):
            image_file_type = image_file_type.lower()
        self.file_type: None | str = image_file_type

        self.stack_images = stack_images

        # image_files are set using create_images.
        self.image_files: None | list[str] | list[ImageFile] = None

        # set object type, object class and object stack class.
        self.object_type = "image"
        self.object_class = ImageFile
        self.object_stack_class = ImageFileStack

    def check(self, raise_error=False):

        if not isinstance(self.image_directory, str):
            if raise_error:
                raise TypeError(
                    f"A directory containing images was expected. Found: an object that was not a path to a "
                    f"directory.")

            return False

        if not os.path.isdir(self.image_directory):
            if raise_error:
                raise TypeError(
                    f"A directory containing images was expected. Found: a path that was not a path to an existing "
                    f"directory: {self.image_directory}")

            return False

        return True

    def create_images(self):

        # Find potentially valid directory paths.
        path_info = list(os.walk(self.image_directory))

        if len(path_info) == 0:
            raise ValueError(
                f"The {self.image_directory} directory is empty, and no {self.object_type}s could be found."
            )

        # Find entries that have associated files.
        path_info = [
            path_info_element for path_info_element in path_info if len(path_info_element[2]) > 0
        ]

        if len(path_info) == 0:
            raise ValueError(
                f"All directories within the {self.image_directory} directory are empty, and no {self.object_type}s "
                f"could be found.")

        # Find entries where the folder structure matches the sub_folder.
        if self.sub_folder is not None:
            path_info = [
                path_info_element for path_info_element in path_info
                if path_info_element[0].endswith(self.sub_folder)
            ]

            if len(path_info) == 0:
                raise ValueError(
                    f"No directories where found in {self.image_directory} that contained the directory "
                    f"substructure {self.sub_folder}.")

        # Add in sample name placeholder to path-information.
        path_info = [list(path_info_element) + [None] for path_info_element in path_info]

        # Find entries that include files of the right file-type. First, we keep only those files that are of the
        # correct file type. Then we filter out entries where no files remain. Note that if file_type is not externally
        # set, all supported image file types are considered.
        allowed_file_extensions = supported_file_types(file_type=self.file_type)

        for ii, path_info_element in enumerate(path_info):
            if len(path_info_element[2]) > 0:
                path_info[ii][2] = [
                    image_file for image_file in path_info_element[2]
                    if image_file.endswith(tuple(allowed_file_extensions))
                ]

        # Find entries that still contain associated files of the right type.
        path_info = [path_info_element for path_info_element in path_info if len(path_info_element[2]) > 0]

        if len(path_info) == 0:
            raise ValueError(
                f"The {self.image_directory} directory (and its subdirectories) did not contain any supported "
                f"{self.object_type} files ({', '.join(allowed_file_extensions)})."
            )

        # Find entries that contain file names that match the image name. First, we keep only those files that contain
        # the file pattern, and then we remove empty directories.
        path_info = self._filter_image_name(path_info=path_info)

        # Use file name pattern to find sample names.
        sample_name_from_pattern = False
        if self.image_name is not None and "#" in self.image_name:
            path_info, sample_name_from_pattern = self._set_pattern_sample_name(path_info=path_info)

            if sample_name_from_pattern:
                path_info = self._filter_pattern_sample_name(path_info=path_info)

        # Use directory name patterns to find sample names (ONLY if sample names are directly provided).
        sample_name_directory = False
        if not sample_name_from_pattern and self.sample_name is not None:
            path_info, sample_name_directory = self._set_directory_sample_name(path_info)

            if sample_name_directory:
                path_info = self._filter_directory_sample_name(path_info)

        # Read and parse image content in subdirectories.
        image_list = []

        for path_info_element in path_info:
            # Make a copy of the object and update the directory and sample name.
            image_sub_directory = copy.deepcopy(self)
            image_sub_directory.image_directory = path_info_element[0]
            image_sub_directory.sample_name = path_info_element[3]
            image_sub_directory.sub_folder = None
            image_sub_directory.image_files = path_info_element[2]

            image_list.append(image_sub_directory._create_images())

        # Flatten list.
        self.image_files = list(chain.from_iterable(image_list))

        # If only a single image is present, and a single sample name provided, and the image does not currently have
        # a sample name: assume that the provided sample name is intended for the particular image.
        if self.image_name is not None and len(self.image_name) == 1 and len(self.image_files) == 1 \
                and self.image_files[0].sample_name is None:
            self.image_files[0].sample_name = self.image_name[0]

        if self.sample_name is not None:
            if len(self.image_files) == 0:
                raise ValueError(
                    f"The {self.image_directory} directory (and its subdirectories) did not contain any"
                    f" {self.object_type}s with the specified sample names ({', '.join(self.sample_name)})."
                )

            # Check for unset sample names.
            if any(current_image.sample_name is None for current_image in self.image_files):

                files_missing_sample_name = [
                    current_image.file_path for current_image in self.image_files
                    if current_image.sample_name is None
                ]

                raise ValueError(
                    f"One or more files ({', '.join(files_missing_sample_name)}) could not be linked to a sample name "
                    f"for checking. You may specify an image file name pattern using the {self.object_type}_name "
                    f"argument, e.g. {self.object_type}_name = '#_*_{self.object_type}' would find John_Doe in "
                    f"John_Doe_CT_{self.object_type}.nii or John_Doe_001_{self.object_type}.nii. Here, '#' indicates "
                    f"the sample name, and '*' is a generic wildcard character."
                )

            # Filter so that only matching sample names remain.
            self.image_files = [
                current_image for current_image in self.image_files
                if current_image.sample_name in self.sample_name
            ]

            if len(self.image_files) == 0:
                raise ValueError(
                    f"The {self.image_directory} directory (and its subdirectories) did not contain any"
                    f" {self.object_type}s with the specified sample names ({', '.join(self.sample_name)})."
                )

            missing_sample_names = set(self.sample_name).difference(
                set(current_image.sample_name for current_image in self.image_files))
            if len(missing_sample_names) > 0:
                raise ValueError(
                    f"The {self.image_directory} directory (and its subdirectories) did not contain all the "
                    f"{self.object_type}s with the required sample names. Missing: {', '.join(missing_sample_names)}"
                )

        if self.modality is not None:
            # Only check images where a modality is present. If all are absent, don't filter on modality,
            if any(current_image.modality is not None for current_image in self.image_files):
                self.image_files = [
                    current_image for current_image in self.image_files
                    if current_image.modality in lookup_modality(self.modality)
                ]

            if len(self.image_files) == 0:
                raise ValueError(
                    f"The {self.image_directory} directory (and its subdirectories) did not contain any "
                    f"{self.object_type}s with the required modality ({self.modality})."
                )

        # Try to stack.
        self.autostack()

    def _filter_image_name(self, path_info):
        allowed_file_extensions = supported_file_types(file_type=self.file_type)
        if self.image_name is not None:
            for ii, path_info_element in enumerate(path_info):
                if len(path_info_element[2]) > 0:
                    path_info[ii][2] = [
                        image_file for image_file in path_info_element[2]
                        if match_file_name(image_file, pattern=self.image_name, file_extension=allowed_file_extensions)
                    ]
            # Find entries that still contain associated image files with the correct name.
            path_info = [path_info_element for path_info_element in path_info if len(path_info_element[2]) > 0]

            if len(path_info) == 0:
                raise ValueError(
                    f"The {self.image_directory} directory (and its subdirectories) did not contain any supported "
                    f"{self.object_type} files ({', '.join(allowed_file_extensions)}) that contain the name pattern "
                    f"({self.image_name}). The name must match exactly. Use wildcard symbol (*) for "
                    f"partial matching, e.g. {'*' + self.image_name[0]}."
                )

        return path_info

    def _set_pattern_sample_name(self, path_info):
        """
        Updates sample names in path_info based on file names. This procedure is EXACT.
        :param path_info: path information as generated by os.walk, with additional sample name element.
        :return: updated path information and update flag.
        """
        allowed_file_extensions = supported_file_types(file_type=self.file_type)
        if self.image_name is None or "#" not in self.image_name:
            return path_info, False

        updated_path_info = []
        for path_info_element in path_info:
            if len(path_info_element[2]) == 0:
                continue

            sample_name = [
                isolate_sample_name(x=file_name, pattern=self.image_name, file_extenstion=allowed_file_extensions)
                for file_name in path_info_element[2]
            ]

            filtered_sample_name = [
                current_sample_name for current_sample_name in sample_name
                if current_sample_name is not None
            ]

            if len(filtered_sample_name) == 0:
                continue

            for current_sample_name in set(filtered_sample_name):
                new_path_info_element = copy.deepcopy(path_info_element)
                new_path_info_element[2] = [
                    file_name for ii, file_name in enumerate(path_info_element[2])
                    if current_sample_name == sample_name[ii]
                ]
                new_path_info_element[3] = current_sample_name

                updated_path_info += [new_path_info_element]

        if len(updated_path_info) == 0:
            raise ValueError(
                f"The file name pattern did not yield any sample names: {self.image_name}"
            )

        return updated_path_info, True

    def _filter_pattern_sample_name(self, path_info):
        # Filter on user-provided sample-names.
        if self.sample_name is not None:
            path_info = [
                path_info_element for path_info_element in path_info if path_info_element[3] in self.sample_name
            ]

            if len(path_info) == 0:
                raise ValueError(
                    f"None of the sample names obtained from file names ("
                    f"{', '.join(set(path_info_element[3] for path_info_element in path_info))}) matched the "
                    f"provided sample names ({', '.join(self.sample_name)})."
                )

        return path_info

    def _set_directory_sample_name(self, path_info):
        """
        Updates sample names in path_info based on directory names. This procedure is EXACT. Sample names are
        inferred if and only if all provided sample names are present in the directory structure.

        :param path_info: path information as generated by os.walk, with additional sample name element.
        :return: updated path information and update flag.
        """

        update_sample_name = False
        if self.sample_name is not None:
            all_samples_selected = True
            sample_name_matches = []
            matching_sample_names = []
            ignore_dirs = [self.image_directory]
            if self.sub_folder is not None:
                ignore_dirs += [self.sub_folder]

            for sample_name in self.sample_name:
                current_sample_matches = [
                    ii for ii in range(len(path_info))
                    if dir_structure_contains_directory(path_info[ii][0], pattern=sample_name, ignore_dir=ignore_dirs)
                ]

                if len(current_sample_matches) > 0:
                    sample_name_matches += current_sample_matches
                    matching_sample_names += [sample_name] * len(current_sample_matches)

                else:
                    all_samples_selected = False
                    break

            # Only filter list if all sample names are uniquely part of their respective paths.
            if all_samples_selected and len(set(sample_name_matches)) == len(sample_name_matches) and len(
                    sample_name_matches) > 0:
                update_sample_name = True

                # Add suggested sample names.
                for ii, path_info_index in enumerate(sample_name_matches):
                    path_info[path_info_index][3] = matching_sample_names[ii]

                # Update path_info
                path_info = [path_info[path_info_index] for path_info_index in sample_name_matches]

        return path_info, update_sample_name

    def _filter_directory_sample_name(self, path_info):
        # Filter on user-provided sample-names.
        if self.sample_name is not None:
            path_info = [
                path_info_element for path_info_element in path_info if path_info_element[3] in self.sample_name
            ]

            if len(path_info) == 0:
                raise ValueError(
                    f"None of the sample names obtained from the directory structure ("
                    f"{', '.join(set(path_info_element[3] for path_info_element in path_info))}) matched the "
                    f"provided sample names ({', '.join(self.sample_name)})."
                )

        return path_info

    def _create_images(self) -> list[ImageFile]:
        """
        Create image objects from files in the directory.

        :return: list of image file objects.
        """

        # Two initial possibilities: the image_files attribute is None, or is set. In case it is None, we need to
        # filter directory contents based on file type and name. Afterwards, the same routine is used.
        if self.image_files is None:

            image_file_list = []

            # Get all contents in the directory, and filter by file type and name.
            dir_contents = os.listdir(self.image_directory)
            allowed_file_extensions = supported_file_types(file_type=self.file_type)

            for current_item in dir_contents:

                # Skip  items that are directories
                if os.path.isdir(os.path.join(self.image_directory, current_item)):
                    continue

                # Skip files that do not have the right extension.
                if not current_item.lower().endswith(tuple(allowed_file_extensions)):
                    continue

                if self.image_name is not None:
                    # Skip files that do not contain the image_name pattern.
                    if not match_file_name(
                            current_item,
                            pattern=self.image_name,
                            file_extension=allowed_file_extensions
                    ):
                        continue

                image_file_list.append(current_item)

            self.image_files = image_file_list

        # From here, the routine is completely the same.
        if len(self.image_files) == 0:
            return []

        image_file_list = []
        for image_file_name in self.image_files:

            # Create image file object.
            image_file = self._create_image_file(image_file_name=image_file_name)
            if image_file is None:
                continue

            if not image_file.check(raise_error=False, remove_metadata=False):
                continue

            image_file.complete()

            image_file_list.append(image_file)

        return image_file_list

    def _create_image_file(self, image_file_name):
        image_file = self.object_class(
            file_path=os.path.join(self.image_directory, image_file_name),
            sample_name=self.sample_name,
            image_modality=self.modality,
            image_file_type=self.file_type
        )

        image_file = image_file.create()

        return image_file

    def autostack(self):
        """
        Form image stacks by combining image file objects of the same class that share identifiers.
        """

        if self.image_files is None or len(self.image_files) == 0:
            return

        if not all(isinstance(image_file_object, ImageFile) for image_file_object in self.image_files):
            raise TypeError("All contents of self.image_files are expected to be ImageFile objects.")

        # Isolate all cases that are not stackable anyway.
        # noinspection PyUnresolvedReferences
        image_file_list = [
            image_file_object for image_file_object in self.image_files
            if not image_file_object.is_stackable(stack_images=self.stack_images)
        ]

        # If none of the images are potentially stackable we don't make any alterations.
        if len(image_file_list) == len(self.image_files):
            return

        image_file_list += list(self._autostack())

        self.image_files = image_file_list

    def _autostack(self):

        # Find stackable objects.
        stackable_image_file_list: list[ImageFile] = [
            image_file_object for image_file_object in self.image_files
            if image_file_object.is_stackable(stack_images=self.stack_images)
        ]

        # Hash identifier data: those files that might be stackable should have the same hash.
        identifier_list = [
            image_file_object.get_identifiers(as_hash=True) for image_file_object in stackable_image_file_list
        ]

        # Create mask to avoid revisiting objects.
        available_objects = [True] * len(identifier_list)

        while any(available_objects):
            # Find next unused identifier.
            identifier = next((
                identifier for ii, identifier in enumerate(identifier_list)
                if available_objects[ii] is True
            ))

            # Identify image file objects that have the same identifier.
            stack_list = [
                image_file_object for ii, image_file_object in enumerate(stackable_image_file_list)
                if available_objects[ii] is True and identifier_list[ii] == identifier
            ]

            # Update availability.
            for ii in range(len(available_objects)):
                if available_objects[ii] is True and identifier_list[ii] == identifier:
                    available_objects[ii] = False

            # If there is only one object to stack, it is not necessary to stack the object.
            if len(stack_list) == 1:
                yield stack_list[0]

            else:
                image_stack = self.object_stack_class(image_file_objects=stack_list).create()
                image_stack.complete()
                yield image_stack


class MaskDirectory(ImageDirectory):

    def __init__(
            self,
            roi_name: None | str | list[str] | dict[str, str] = None,
            **kwargs
    ):
        super().__init__(**kwargs)

        # set object type, object class and object stack class.
        self.object_type = "mask"
        self.object_class = MaskFile
        self.object_stack_class = MaskFileStack

        # Set roi name
        self.roi_name = roi_name

    def _create_image_file(self, image_file_name):
        image_file = self.object_class(
            file_path=os.path.join(self.image_directory, image_file_name),
            sample_name=self.sample_name,
            image_modality=self.modality,
            image_file_type=self.file_type,
            roi_name=self.roi_name
        )

        image_file = image_file.create()

        return image_file



from functools import lru_cache
from typing import Generator

import numpy as np

from mirp._images.generic_image import GenericImage
from mirp._masks.base_mask import BaseMask
from mirp._features.base_feature import Feature
from mirp.settings.feature_parameters import FeatureExtractionSettingsClass


class DataStatistics(object):

    def __init__(self):
        # Raw intensity data
        self.image: np.ndarray | None = None

        # Number of voxels
        self.n_voxels: int | None = None

        # Mean intensity
        self.mu: float | None = None

        # Intensity standard deviation
        self.sigma: float | None = None

    def compute(self, image: GenericImage, mask: BaseMask):
        # Skip processing if input image and/or roi are missing
        if image is None or mask is None:
            return

        # Check if data actually exists
        if image.is_empty() or mask.roi_intensity.is_empty_mask():
            return

        # Convert to dataframe and remove entries outside the mask.
        image = mask.as_pandas_dataframe(image=image, intensity_mask=True)
        image = image[image.roi_int_mask == True]
        if len(image) == 0:
            return

        # Set image data
        self.image = image.g.values

        # Number of voxels
        self.n_voxels = len(image)

        # Mean intensity
        self.mu = np.mean(self.image)

        # Standard deviation
        self.sigma = np.std(self.image, ddof=0)

    def is_empty(self):
        return self.image is None


class FeatureStat(Feature):

    def __init__(
            self,
            **kwargs
    ):
        super().__init__(**kwargs)

    @staticmethod
    @lru_cache(maxsize=1)
    def _get_data(
            image: GenericImage,
            mask: BaseMask
    ) -> DataStatistics:
        data = DataStatistics()
        data.compute(image=image, mask=mask)

        return data

    def clear_local_cache(self, other):
        if not isinstance(other, FeatureStat):
            self._get_data.cache_clear()

    def clear_cache(self):
        super().clear_cache()
        self._get_data.cache_clear()

    def compute(self, image: GenericImage, mask: BaseMask):
        # Get data.
        data = self._get_data(image=image, mask=mask)

        # Compute feature value.
        if data.is_empty():
            self.value = np.nan
        else:
            self.value = self._compute(data=data)

    @staticmethod
    def _compute(data: DataStatistics):
        raise NotImplementedError("Implement _compute for feature-specific computation.")

    def create_table_name(self):
        table_elements = self._get_base_table_name_element()
        self.table_name = "_".join(table_elements)


class FeatureStatMean(FeatureStat):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "Statistics - mean"
        self.abbr_name = "stat_mean"
        self.ibsi_id = "Q4LE"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(data: DataStatistics) -> float:
        return data.mu


class FeatureStatVariance(FeatureStat):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "Statistics - variance"
        self.abbr_name = "stat_var"
        self.ibsi_id = "ECT3"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(data: DataStatistics) -> float:
        return data.sigma ** 2.0


class FeatureStatSkewness(FeatureStat):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "Statistics - skewness"
        self.abbr_name = "stat_skew"
        self.ibsi_id = "KE2A"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(data: DataStatistics) -> float:
        if data.sigma == 0.0:
            return 0.0
        return np.sum((data.image - data.mu) ** 3.0) / (data.n_voxels * data.sigma ** 3.0)


class FeatureStatKurtosis(FeatureStat):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "Statistics - kurtosis"
        self.abbr_name = "stat_kurt"
        self.ibsi_id = "IPH6"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(data: DataStatistics) -> float:
        if data.sigma == 0.0:
            return 0.0
        return np.sum((data.image - data.mu) ** 4.0) / (data.n_voxels * data.sigma ** 4.0) - 3.0


class FeatureStatMedian(FeatureStat):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "Statistics - median"
        self.abbr_name = "stat_median"
        self.ibsi_id = "Y12H"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(data: DataStatistics) -> float:
        return np.median(data.image)


class FeatureStatMinimum(FeatureStat):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "Statistics - minimum"
        self.abbr_name = "stat_min"
        self.ibsi_id = "1GSF"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(data: DataStatistics) -> float:
        return np.min(data.image)


class FeatureStatPercentile(FeatureStat):
    def __init__(self, percentile: float, **kwargs):
        # This class allows for passing percentile values.
        super().__init__(**kwargs)

        # Set percentile
        self.percentile = percentile

        # Set IBSI identifier. Only the 10th and 90th percentile are explicitly identified.
        ibsi_id = ""
        if percentile == 10.0:
            ibsi_id = "QG58"
        elif percentile == 90.0:
            ibsi_id = "8DWT"
        self.ibsi_id = ibsi_id

        if percentile.is_integer():
            self.name = f"Statistics - {int(percentile)}th percentile"
            self.abbr_name = f"stat_p{int(percentile)}"
        else:
            self.name = self.name = f"Statistics - {percentile}th percentile"
            self.abbr_name = f"stat_p{percentile}"

        self.ibsi_compliant = True

    def compute(self, image: GenericImage, mask: BaseMask):
        # Get data.
        data = self._get_data(image=image, mask=mask)

        # Compute feature value.
        if data.is_empty():
            self.value = np.nan
        else:
            self.value = np.percentile(data.image, q=self.percentile)


class FeatureStatMaximum(FeatureStat):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "Statistics - maximum"
        self.abbr_name = "stat_max"
        self.ibsi_id = "84IY"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(data: DataStatistics) -> float:
        return np.max(data.image)


class FeatureStatInterQuartileRange(FeatureStat):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "Statistics - interquartile range"
        self.abbr_name = "stat_iqr"
        self.ibsi_id = "SALO"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(data: DataStatistics) -> float:
        return np.percentile(data.image, q=75) - np.percentile(data.image, q=25)


class FeatureStatRange(FeatureStat):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "Statistics - range"
        self.abbr_name = "stat_range"
        self.ibsi_id = "2OJQ"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(data: DataStatistics) -> float:
        return float(np.max(data.image) - np.min(data.image))


class FeatureStatMeanAbsoluteDeviation(FeatureStat):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "Statistics - mean absolute deviation"
        self.abbr_name = "stat_mad"
        self.ibsi_id = "4FUA"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(data: DataStatistics) -> float:
        return np.mean(np.abs(data.image - data.mu))


class FeatureStatRobustMeanAbsoluteDeviation(FeatureStat):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "Statistics - robust mean absolute deviation"
        self.abbr_name = "stat_rmad"
        self.ibsi_id = "1128"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(data: DataStatistics) -> float:
        selected_values = data.image[
            (data.image >= np.percentile(data.image, q=10))
            & (data.image <= np.percentile(data.image, q=90))
        ]
        return np.mean(np.abs(selected_values - np.mean(selected_values)))


class FeatureStatMedianAbsoluteDeviation(FeatureStat):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "Statistics - median absolute deviation"
        self.abbr_name = "stat_medad"
        self.ibsi_id = "N72L"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(data: DataStatistics) -> float:
        return np.mean(np.abs(data.image - np.median(data.image)))


class FeatureStatCoefficientOfVariation(FeatureStat):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "Statistics - coefficient of variation"
        self.abbr_name = "stat_cov"
        self.ibsi_id = "7TET"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(data: DataStatistics) -> float:
        if data.sigma == 0.0:
            return 0.0
        return data.sigma / data.mu


class FeatureStatQuartileCoefficientOfDispersion(FeatureStat):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "Statistics - quartile coefficient of dispersion"
        self.abbr_name = "stat_qcod"
        self.ibsi_id = "9S40"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(data: DataStatistics) -> float:
        denom = np.percentile(data.image, q=75) + np.percentile(data.image, q=25)
        if denom == 0.0:
            return 1.0

        return (np.percentile(data.image, q=75) - np.percentile(data.image, q=25)) / denom


class FeatureStatEnergy(FeatureStat):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "Statistics - energy"
        self.abbr_name = "stat_energy"
        self.ibsi_id = "N8CA"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(data: DataStatistics) -> float:
        return np.sum(data.image ** 2.0)


class FeatureStatRootMeanSquare(FeatureStat):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "Statistics - "
        self.abbr_name = "stat_rms"
        self.ibsi_id = "5ZWQ"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(data: DataStatistics) -> float:
        return  np.sqrt(np.sum(data.image ** 2.0) / data.n_voxels)


def get_statistics_class_dict() -> dict[str, FeatureStat]:
    class_dict = {
        "stat_mean": FeatureStatMean,
        "stat_var": FeatureStatVariance,
        "stat_skew": FeatureStatSkewness,
        "stat_kurt": FeatureStatKurtosis,
        "stat_median": FeatureStatMedian,
        "stat_min": FeatureStatMinimum,
        "stat_p": FeatureStatPercentile,
        "stat_max": FeatureStatMaximum,
        "stat_iqr": FeatureStatInterQuartileRange,
        "stat_range": FeatureStatRange,
        "stat_mad": FeatureStatMeanAbsoluteDeviation,
        "stat_rmad": FeatureStatRobustMeanAbsoluteDeviation,
        "stat_medad": FeatureStatMedianAbsoluteDeviation,
        "stat_cov": FeatureStatCoefficientOfVariation,
        "stat_qcod": FeatureStatQuartileCoefficientOfDispersion,
        "stat_energy": FeatureStatEnergy,
        "stat_rms": FeatureStatRootMeanSquare
    }

    return class_dict


def generate_stat_features(
        settings: FeatureExtractionSettingsClass,
        features: None | list[str] = None
) -> Generator[FeatureStat, None, None]:
    class_dict = get_statistics_class_dict()
    stat_features = list(class_dict.keys())

    # Populate features if available.
    if features is None and settings.has_stats_family():
        features = stat_features

    # Terminate early if no features are set, and none are required.
    if features is None:
        return

    # Select only statistical features, and return if none are present.
    features = [feature for feature in features if feature in stat_features]
    if len(features) == 0:
        return

    # Set default percentiles.
    percentiles = [10.0, 90.0]

    for feature in features:
        if feature == "stat_p":
            for percentile in percentiles:
                yield class_dict[feature](
                    percentile=percentile
                )
        else:
            yield class_dict[feature]()



import warnings
from dataclasses import dataclass
from typing import Any
from mirp.settings.utilities import setting_def


@dataclass
class GeneralSettingsClass:
    """
    Set of overall process parameters. The most important parameter here is ``by_slice`` which affects how images are
    processed and features are computed.

    Parameters
    ----------
    by_slice: bool, optional, default: False
        Defines whether image processing and computations should be performed in 2D (True) or 3D (False).

    ibsi_compliant: bool, optional, default: True
        Limits use of methods and computation of features to those that exist in the IBSI reference standard.

    mask_merge: bool, optional, default: False
        Defines whether multiple mask objects should be combined into a single mask.

    mask_split: bool, optional, default: False
        Defines whether a mask that contains multiple regions should be split into separate mask objects.

    mask_select_largest_region: bool, optional, default: False
        Defines whether the largest region within a mask object should be selected. For example, in a mask that
        contains multiple separate lesions. ``mask_select_largest_region = True`` will remove all but the largest
        lesion.

    mask_select_largest_slice: bool, optional, default: False
        Defines whether the largest slice within a mask object should be selected.

    config_str: str, optional
        Sets a configuration string, which can be used to differentiate results obtained using other settings.

    no_approximation: bool, optional, default: False
        Disables approximation within MIRP. This currently only affects computation of features such as Geary's
        c-measure. Can be True or False (default). False means that approximation is performed.

    **kwargs: dict, optional
        Unused keyword arguments.
    """

    def __init__(
            self,
            by_slice: bool = False,
            ibsi_compliant: bool = True,
            mask_merge: bool = False,
            mask_split: bool = False,
            mask_select_largest_region: bool = False,
            mask_select_largest_slice: bool = False,
            config_str: str = "",
            no_approximation: bool = False,
            **kwargs
    ):

        if not isinstance(by_slice, bool):
            raise ValueError("The by_slice parameter should be a boolean.")

        # Set by_slice and select_slice parameters.
        self.by_slice: bool = by_slice

        # Set IBSI-compliance flag.
        self.ibsi_compliant: bool = ibsi_compliant

        self.mask_merge = mask_merge
        self.mask_split = mask_split
        self.mask_select_largest_region = mask_select_largest_region

        if mask_select_largest_slice and not by_slice:
            warnings.warn("A 2D approach is used as the largest slice is selected.", UserWarning)
            self.by_slice = True

        self.mask_select_largest_slice = mask_select_largest_slice

        # Set configuration string.
        self.config_str: str = config_str

        # Set approximation of features.
        self.no_approximation: bool = no_approximation


def get_general_settings() -> list[dict[str, Any]]:
    return [
        setting_def("by_slice", "bool", test=True),
        setting_def("ibsi_compliant", "bool", test=True),
        setting_def("mask_merge", "bool", test=True),
        setting_def("mask_split", "bool", test=True),
        setting_def("mask_select_largest_region", "bool", test=True),
        setting_def("mask_select_largest_slice", "bool", test=True),
        setting_def("config_str", "str", test="test_config"),
        setting_def("no_approximation", "bool", test=True)
    ]
